# creates: acf_example.png
import numpy as np
import matplotlib.pyplot as plt
from gpaw.core import UniformGrid


alpha = 4.0
rcut = 2.0
l = 0
gauss = (l, rcut, lambda r: (4 * np.pi)**0.5 * np.exp(-alpha * r**2))
grid = UniformGrid(cell=[4.0, 2.0, 2.0], size=[40, 20, 20])
pos = [[0.25, 0.5, 0.5], [0.75, 0.5, 0.5]]
acf_aR = grid.atom_centered_functions([[gauss], [gauss]], pos)
coef_as = acf_aR.empty(dims=(2,))
coef_as[0] = [[1], [-1]]
coef_as[1] = [[2], [1]]
print(coef_as.data, coef_as[0])
f_sR = grid.zeros(2)
acf_aR.add_to(f_sR, coef_as)
x = grid.xyz()[:, 10, 10, 0]
y1, y2 = f_sR.data[:, :, 10, 10]
ax = plt.subplot(1, 1, 1)
ax.plot(x, y1, 'o-')
ax.plot(x, y2, 'x-')
# plt.show()
plt.savefig('acf_example.png')



"""
Exponential Transformation Direct Minimization
"""

from gpaw.xc import xc_string_to_dict
from gpaw.directmin.sd_etdm import SteepestDescent, FRcg, QuickMin, LBFGS, \
    LBFGS_P, LSR1P
from gpaw.directmin.ls_etdm import MaxStep, StrongWolfeConditions, Parabola


def search_direction(method):
    if isinstance(method, str):
        method = xc_string_to_dict(method)

    if isinstance(method, dict):
        kwargs = method.copy()
        name = kwargs.pop('name')

        searchdir = {'sd': SteepestDescent,
                     'fr-cg': FRcg,
                     'quick-min': QuickMin,
                     'l-bfgs': LBFGS,
                     'l-bfgs-p': LBFGS_P,
                     'l-sr1p': LSR1P
                     }[name](**kwargs)

        return searchdir
    else:
        raise ValueError('Check keyword for search direction!')


def line_search_algorithm(method, objective_function, searchdir_algo):
    if isinstance(method, str):
        method = xc_string_to_dict(method)

    if isinstance(method, dict):
        kwargs = method.copy()
        name = kwargs.pop('name')
        if name == 'swc-awc':
            # for swc-awc we need to know
            # what search. dir. algo is used
            if 'searchdirtype' not in kwargs:
                kwargs['searchdirtype'] = searchdir_algo.type

        ls_algo = {'max-step': MaxStep,
                   'parabola': Parabola,
                   'swc-awc': StrongWolfeConditions
                   }[name](objective_function, **kwargs)

        return ls_algo
    else:
        raise ValueError('Check keyword for line search!')



"""AFM NiO: Make sure spin-polarized GW works."""
from gpaw.response.g0w0 import G0W0

gw = G0W0('nio.gpw',
          'gw6ok',
          nbands=100,
          ecut=100,
          truncation='wigner-seitz',
          kpts=[(0, 0, 0)],
          relbands=(-1, 1),
          nblocks=6)
result = gw.calculate()

# Make sure gaps in both spin-channels are the same:
pbe_sn = result['eps'][:, 0]
pbe_gap_s = pbe_sn[:, 1] - pbe_sn[:, 0]
assert abs(pbe_gap_s - 2.319).max() < 0.01
qp_sn = result['qp'][:, 0]
qp_gap_s = qp_sn[:, 1] - qp_sn[:, 0]
assert abs(qp_gap_s - 3.743).max() < 0.01



from ase.build import molecule
from gpaw import GPAW

a = 8.0
h = 0.2

energies = {}
with open(f'results-{h:.2f}.txt', 'w') as resultfile:

    for name in ['H2O', 'H', 'O']:
        system = molecule(name)
        system.set_cell((a, a, a))
        system.center()
    
        calc = GPAW(h=h,
                    txt=f'gpaw-{name}-{h:.2f}.txt')
        if name == 'H' or name == 'O':
            calc.set(hund=True)
    
        system.calc = calc
    
        energy = system.get_potential_energy()
        energies[name] = energy
        print(name, energy, file=resultfile)
    
    e_atomization = energies['H2O'] - 2 * energies['H'] - energies['O']
    print(e_atomization, file=resultfile)



from __future__ import annotations

from math import pi

import _gpaw
import numpy as np

import gpaw.fftw as fftw
from gpaw.core.arrays import DistributedArrays
from gpaw.core.domain import Domain
from gpaw.core.matrix import Matrix
from gpaw.core.pwacf import PlaneWaveAtomCenteredFunctions
from gpaw.core.uniform_grid import UniformGrid, UniformGridFunctions
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new import prod, zip
from gpaw.pw.descriptor import pad
from gpaw.typing import (Array1D, Array2D, Array3D, ArrayLike1D, ArrayLike2D,
                         Vector)


class PlaneWaves(Domain):
    def __init__(self,
                 *,
                 ecut: float,
                 cell: ArrayLike1D | ArrayLike2D,
                 kpt: Vector = None,
                 comm: MPIComm = serial_comm,
                 dtype=None):
        """Description of plane-wave basis.

        parameters
        ----------
        ecut:
            Cutoff energy for kinetic energy of plane waves.
        cell:
            Unit cell given as three floats (orthorhombic grid), six floats
            (three lengths and the angles in degrees) or a 3x3 matrix.
        comm:
            Communicator for distribution of plane-waves.
        kpt:
            K-point for Block-boundary conditions specified in units of the
            reciprocal cell.
        dtype:
            Data-type (float or complex).
        """
        self.ecut = ecut
        Domain.__init__(self, cell, (True, True, True), kpt, comm, dtype)

        G_plus_k_Gv, ekin_G, self.indices_cG = find_reciprocal_vectors(
            ecut, self.cell_cv, self.kpt_c, self.dtype)

        # Find distribution:
        S = comm.size
        ng = len(ekin_G)
        self.maxmysize = (ng + S - 1) // S
        ng1 = comm.rank * self.maxmysize
        ng2 = min(ng1 + self.maxmysize, ng)
        self.ng1 = ng1
        self.ng2 = ng2

        # Distribute things:
        self.ekin_G = ekin_G[ng1:ng2].copy()
        self.ekin_G.flags.writeable = False
        # self.myindices_cG = self.indices_cG[:, ng1:ng2]
        self.G_plus_k_Gv = G_plus_k_Gv[ng1:ng2].copy()

        self.shape = (ng,)
        self.myshape = (len(self.ekin_G),)

        self.dv = abs(np.linalg.det(self.cell_cv))

        self._indices_cache: dict[tuple[int, ...], Array1D] = {}

        self.qspiral_v = None

    def __repr__(self) -> str:
        m = self.myshape[0]
        n = self.shape[0]
        r = Domain.__repr__(self).replace(
            'Domain(',
            f'PlaneWaves(ecut={self.ecut} <coefs={m}/{n}>, ')
        if self.qspiral_v is None:
            return r
        q = self.cell_cv @ self.qspiral_v / (2 * pi)
        return f'{r[:-1]}, qsiral={q}'

    def global_shape(self) -> tuple[int, ...]:
        """Tuple with one element: number of plane waves."""
        return self.shape

    def reciprocal_vectors(self) -> Array2D:
        """Returns reciprocal lattice vectors, G + k, in xyz coordinates."""
        return self.G_plus_k_Gv

    def kinetic_energies(self) -> Array1D:
        """Kinetic energy of plane waves.

        :::

             _ _ 2
            |G+k| / 2

        """
        return self.ekin_G

    def empty(self,
              dims: int | tuple[int, ...] = (),
              comm: MPIComm = serial_comm) -> PlaneWaveExpansions:
        """Create new PlaneWaveExpanions object.

        parameters
        ----------
        dims:
            Extra dimensions.
        comm:
            Distribute dimensions along this communicator.
        """
        return PlaneWaveExpansions(self, dims, comm)

    def new(self,
            *,
            ecut: float = None,
            kpt=None,
            comm: MPIComm | str = 'inherit') -> PlaneWaves:
        """Create new plane-wave expansion description."""
        comm = self.comm if comm == 'inherit' else comm
        return PlaneWaves(ecut=ecut or self.ecut,
                          cell=self.cell_cv,
                          kpt=self.kpt_c if kpt is None else kpt,
                          dtype=self.dtype,
                          comm=comm or serial_comm)

    def indices(self, shape: tuple[int, ...]) -> Array1D:
        """Return indices into FFT-grid."""
        Q_G = self._indices_cache.get(shape)
        if Q_G is None:
            Q_G = np.ravel_multi_index(self.indices_cG, shape,  # type: ignore
                                       mode='wrap').astype(np.int32)
            self._indices_cache[shape] = Q_G
        return Q_G

    def cut(self, array_Q: Array3D) -> Array1D:
        """Cut out G-vectors with (G+k)^2/2<E_kin."""
        return array_Q.ravel()[self.indices(array_Q.shape)]

    def paste(self, coef_G: Array1D, array_Q: Array3D) -> None:
        """Paste G-vectors with (G+k)^2/2<E_kin into 3-D FFT grid and
        zero-pad."""
        Q_G = self.indices(array_Q.shape)
        # array_Q[:] = 0.0
        # array_Q.ravel()[Q_G] = coef_G
        _gpaw.pw_insert(coef_G, Q_G, 1.0, array_Q)

    def map_indices(self, other: PlaneWaves) -> tuple[Array1D, list[Array1D]]:
        """Map from one (distributed) set of plane waves to smaller global set.

        Say we have 9 G-vector on two cores::

           5 3 4             . 3 4           0 . .
           2 0 1 -> rank=0:  2 0 1  rank=1:  . . .
           8 6 7             . . .           3 1 2

        and we want a mapping to these 5 G-vectors::

             3
           2 0 1
             4

        On rank=0: the return values are::

           [0, 1, 2, 3], [[0, 1, 2, 3], [4]]

        and for rank=1::

           [1], [[0, 1, 2, 3], [4]]
        """
        size_c = tuple(self.indices_cG.ptp(axis=1) + 1)  # type: ignore
        Q_G = self.indices(size_c)
        G_Q = np.empty(prod(size_c), int)
        G_Q[Q_G] = np.arange(len(Q_G))
        G_g = G_Q[other.indices(size_c)]
        ng1 = 0
        g_r = []
        for rank in range(self.comm.size):
            ng2 = min(ng1 + self.maxmysize, self.shape[0])
            myg = (ng1 <= G_g) & (G_g < ng2)
            g_r.append(np.nonzero(myg)[0])
            if rank == self.comm.rank:
                my_G_g = G_g[myg] - ng1
            ng1 = ng2
        return my_G_g, g_r

    def atom_centered_functions(self,
                                functions,
                                positions,
                                *,
                                atomdist=None,
                                integral=None,
                                cut=False):
        """Create PlaneWaveAtomCenteredFunctions object."""
        if self.qspiral_v is None:
            return PlaneWaveAtomCenteredFunctions(functions, positions, self,
                                                  atomdist=atomdist)

        from gpaw.new.spinspiral import SpiralPWACF
        return SpiralPWACF(functions, positions, self,
                           atomdist=atomdist,
                           qspiral_v=self.qspiral_v)


class PlaneWaveExpansions(DistributedArrays[PlaneWaves]):
    def __init__(self,
                 pw: PlaneWaves,
                 dims: int | tuple[int, ...] = (),
                 comm: MPIComm = serial_comm,
                 data: np.ndarray = None):
        """Object for storing function(s) as a plane-wave expansions.

        parameters
        ----------
        pw:
            Description of plane-waves.
        dims:
            Extra dimensions.
        comm:
            Distribute plane-waves along this communicator.
        data:
            Data array for storage.
        """
        DistributedArrays. __init__(self, dims, pw.myshape,
                                    comm, pw.comm,
                                    data, pw.dv, complex)
        self.desc = pw
        self._matrix: Matrix | None

    def __repr__(self):
        txt = f'PlaneWaveExpansions(pw={self.desc}, dims={self.dims}'
        if self.comm.size > 1:
            txt += f', comm={self.comm.rank}/{self.comm.size}'
        return txt + ')'

    def __getitem__(self, index: int) -> PlaneWaveExpansions:
        data = self.data[index]
        return PlaneWaveExpansions(self.desc, data.shape[:-1], data=data)

    def __iter__(self):
        for data in self.data:
            yield PlaneWaveExpansions(self.desc, data.shape[:-1], data=data)

    def new(self, data=None):
        """Create new PlaneWaveExpansions object of same kind.

        Parameters
        ----------
        data:
            Array to use for storage.
        """
        if data is None:
            data = np.empty_like(self.data)
        else:
            # Number of plane-waves depends on the k-point.  We therfore
            # allow for data to be bigger than needed:
            data = data.ravel()[:self.data.size].reshape(self.data.shape)
        return PlaneWaveExpansions(self.desc, self.dims, self.comm, data)

    def copy(self):
        """Create a copy (surprise!)."""
        a = self.new()
        a.data[:] = self.data
        return a

    def _arrays(self):
        shape = self.data.shape
        return self.data.reshape((prod(shape[:-1]), shape[-1]))

    @property
    def matrix(self) -> Matrix:
        """Matrix view of data."""
        if self._matrix is not None:
            return self._matrix

        shape = (self.dims[0], prod(self.dims[1:]) * self.myshape[0])
        myshape = (self.mydims[0], prod(self.mydims[1:]) * self.myshape[0])
        dist = (self.comm, -1, 1)
        data = self.data.reshape(myshape)

        if self.desc.dtype == float:
            data = data.view(float)
            shape = (shape[0], shape[1] * 2)

        self._matrix = Matrix(*shape, data=data, dist=dist)
        return self._matrix

    def ifft(self, *, plan=None, grid=None, out=None, periodic=False):
        """Do inverse FFT to uniform grid.

        Parameters
        ----------
        plan:
            Plan for inverse FFT.
        grid:
            Target grid.
        out:
            Target UniformGridFunctions object.
        """
        comm = self.desc.comm
        if out is None:
            out = grid.empty(self.dims)
        assert self.desc.dtype == out.desc.dtype
        assert out.desc.pbc_c.all()
        assert comm.size == out.desc.comm.size

        this = self.gather()
        if this is not None:
            plan = plan or out.desc.fft_plans()
            for coef_G, out1 in zip(this._arrays(), out.flat()):
                self.desc.paste(coef_G, plan.tmp_Q)
                if self.desc.dtype == float:
                    t = plan.tmp_Q[:, :, 0]
                    n, m = (s // 2 - 1 for s in out.desc.size_c[:2])
                    t[0, -m:] = t[0, m:0:-1].conj()
                    t[n:0:-1, -m:] = t[-n:, m:0:-1].conj()
                    t[-n:, -m:] = t[n:0:-1, m:0:-1].conj()
                    t[-n:, 0] = t[n:0:-1, 0].conj()
                plan.ifft()
                out1.scatter_from(plan.tmp_R)
        else:
            for out1 in out.flat():
                out1.scatter_from(None)

        if not periodic:
            out.multiply_by_eikr()

        return out

    def interpolate(self,
                    plan1: fftw.FFTPlans = None,
                    plan2: fftw.FFTPlans = None,
                    grid: UniformGrid = None,
                    out: UniformGridFunctions = None) -> UniformGridFunctions:
        assert plan1 is None
        return self.ifft(plan=plan2, grid=grid, out=out)

    def gather(self, out=None, broadcast=False):
        """Gather coefficients on master."""
        comm = self.desc.comm

        if comm.size == 1:
            if out is None:
                return self
            out.data[:] = self.data
            return out

        if out is None:
            if comm.rank == 0 or broadcast:
                pw = self.desc.new(comm=serial_comm)
                out = pw.empty(self.dims)
            else:
                out = Empty(self.dims)

        if comm.rank == 0:
            data = np.empty(self.desc.maxmysize * comm.size, complex)
        else:
            data = None

        for input, output in zip(self._arrays(), out._arrays()):
            mydata = pad(input, self.desc.maxmysize)
            comm.gather(mydata, 0, data)
            if comm.rank == 0:
                output[:] = data[:len(output)]

        if broadcast:
            comm.broadcast(out.data, 0)

        return out if not isinstance(out, Empty) else None

    def scatter_from(self, data: Array1D = None) -> None:
        """Scatter data from rank-0 to all ranks."""
        comm = self.desc.comm
        if comm.size == 1:
            self.data[:] = data
            return

        assert self.dims == ()

        if comm.rank == 0:
            data = pad(data, comm.size * self.desc.maxmysize)
            comm.scatter(data, self.data, 0)
        else:
            buf = np.empty(self.desc.maxmysize, complex)
            comm.scatter(None, buf, 0)
            self.data[:] = buf[:len(self.data)]

    def integrate(self, other: PlaneWaveExpansions = None) -> np.ndarray:
        """Integral of self or self time cc(other)."""
        if other is not None:
            assert self.comm.size == 1
            assert self.desc.dtype == other.desc.dtype
            a = self._arrays()
            b = other._arrays()
            dv = self.dv
            if self.desc.dtype == float:
                a = a.view(float)
                b = b.view(float)
                dv *= 2
            result = a @ b.T.conj()
            if self.desc.dtype == float and self.desc.comm.rank == 0:
                result -= 0.5 * np.outer(a[:, 0], b[:, 0])
            self.desc.comm.sum(result)
            result.shape = self.dims + other.dims
        else:
            dv = self.dv
            if self.desc.comm.rank == 0:
                result = self.data[..., 0]
            else:
                result = np.empty(self.mydims, complex)
            self.desc.comm.broadcast(result, 0)

        if self.desc.dtype == float:
            result = result.real
        return result * dv

    def _matrix_elements_correction(self,
                                    M1: Matrix,
                                    M2: Matrix,
                                    out: Matrix,
                                    symmetric: bool) -> None:
        if self.desc.dtype == float:
            out.data *= 2.0
            if self.desc.comm.rank == 0:
                correction = np.outer(M1.data[:, 0],
                                      M2.data[:, 0]) * self.dv
                if symmetric:
                    correction *= 0.5
                    out.data -= correction
                    out.data -= correction.T
                else:
                    out.data -= correction

    def norm2(self, kind: str = 'normal') -> np.ndarray:
        r"""Calculate integral over cell.

        For kind='normal' we calculate:::

          /   _  2 _   --    2
          ||a(r)| dr = > |c | V,
          /            --  G
                        G

        where V is the volume of the unit cell.

        And for kind='kinetic':::

           1  --    2  2
          --- > |c |  G V,
           2  --  G
               G

        """
        a_xG = self._arrays().view(float)
        if kind == 'normal':
            result_x = np.einsum('xG, xG -> x', a_xG, a_xG)
        elif kind == 'kinetic':
            a_xG.shape = (len(a_xG), -1, 2)
            result_x = np.einsum('xGi, xGi, G -> x',
                                 a_xG, a_xG, self.desc.ekin_G)
        else:
            1 / 0
        if self.desc.dtype == float:
            result_x *= 2
            if self.desc.comm.rank == 0 and kind == 'normal':
                result_x -= a_xG[:, 0]**2
        self.desc.comm.sum(result_x)
        result_x.shape = self.mydims
        return result_x * self.dv

    def abs_square(self,
                   weights: Array1D,
                   out: UniformGridFunctions = None) -> None:
        """Add weighted absolute square of data to output array."""
        assert out is not None
        tmp_R = out.desc.new(dtype=self.desc.dtype).empty()
        for f, psit_G in zip(weights, self):
            # Same as (but much faster):
            # out.data += f * abs(psit.ifft().data)**2
            psit_G.ifft(out=tmp_R)
            _gpaw.add_to_density(f, tmp_R.data, out.data)

    def to_pbc_grid(self):
        return self

    def randomize(self) -> None:
        """Insert random numbers between -0.5 and 0.5 into data."""
        seed = [self.comm.rank, self.desc.comm.rank]
        rng = np.random.default_rng(seed)
        a = self.data.view(float)
        rng.random(a.shape, out=a)
        a -= 0.5


class Empty:
    def __init__(self, dims):
        self.dims = dims

    def _arrays(self):
        for _ in range(prod(self.dims)):
            yield


def find_reciprocal_vectors(ecut: float,
                            cell: Array2D,
                            kpt=np.zeros(3),
                            dtype=complex) -> tuple[Array2D,
                                                    Array1D,
                                                    Array2D]:
    """Find reciprocal lattice vectors inside sphere.

    >>> cell = np.eye(3)
    >>> ecut = 0.5 * (2 * pi)**2
    >>> G, e, i = find_reciprocal_vectors(ecut, cell)
    >>> G
    array([[ 0.        ,  0.        ,  0.        ],
           [ 0.        ,  0.        ,  6.28318531],
           [ 0.        ,  0.        , -6.28318531],
           [ 0.        ,  6.28318531,  0.        ],
           [ 0.        , -6.28318531,  0.        ],
           [ 6.28318531,  0.        ,  0.        ],
           [-6.28318531,  0.        ,  0.        ]])
    >>> e
    array([ 0.       , 19.7392088, 19.7392088, 19.7392088, 19.7392088,
           19.7392088, 19.7392088])
    >>> i
    array([[ 0,  0,  0,  0,  0,  1, -1],
           [ 0,  0,  0,  1, -1,  0,  0],
           [ 0,  1, -1,  0,  0,  0,  0]])
    """
    Gcut = (2 * ecut)**0.5
    n = Gcut * (cell**2).sum(axis=1)**0.5 / (2 * pi) + abs(kpt)
    size = 2 * n.astype(int) + 4

    if dtype == float:
        size[2] = size[2] // 2 + 1
        i_Qc = np.indices(size).transpose((1, 2, 3, 0))
        i_Qc[..., :2] += size[:2] // 2
        i_Qc[..., :2] %= size[:2]
        i_Qc[..., :2] -= size[:2] // 2
    else:
        i_Qc = np.indices(size).transpose((1, 2, 3, 0))  # type: ignore
        half = [s // 2 for s in size]
        i_Qc += half
        i_Qc %= size
        i_Qc -= half

    # Calculate reciprocal lattice vectors:
    B_cv = 2.0 * pi * np.linalg.inv(cell).T
    # i_Qc.shape = (-1, 3)
    G_plus_k_Qv = (i_Qc + kpt) @ B_cv

    ekin = 0.5 * (G_plus_k_Qv**2).sum(axis=3)
    mask = ekin <= ecut

    assert not mask[size[0] // 2].any()
    assert not mask[:, size[1] // 2].any()
    if dtype == complex:
        assert not mask[:, :, size[2] // 2].any()
    else:
        assert not mask[:, :, -1].any()

    if dtype == float:
        mask &= ((i_Qc[..., 2] > 0) |
                 (i_Qc[..., 1] > 0) |
                 ((i_Qc[..., 0] >= 0) & (i_Qc[..., 1] == 0)))

    indices = i_Qc[mask]
    ekin = ekin[mask]
    G_plus_k = G_plus_k_Qv[mask]

    return G_plus_k, ekin, indices.T


x = '''
class PWMapping:
    def __init__(self, pw1: PlaneWaves, pw2: PlaneWaves):
        """Mapping from pd1 to pd2."""
        N_c = pw1.grid.size
        N2_c = pw2.grid.size
        assert pw1.grid.dtype == pw2.grid.dtype
        if pw1.grid.dtype == float:
            N_c = N_c.copy()
            N_c[2] = N_c[2] // 2 + 1
            N2_c = N2_c.copy()
            N2_c[2] = N2_c[2] // 2 + 1

        Q1_G = pw1.myindices
        Q1_Gc = np.empty((len(Q1_G), 3), int)
        Q1_Gc[:, 0], r_G = divmod(Q1_G, N_c[1] * N_c[2])
        Q1_Gc.T[1:] = divmod(r_G, N_c[2])
        if pw1.grid.dtype == float:
            C = 2
        else:
            C = 3
        Q1_Gc[:, :C] += N_c[:C] // 2
        Q1_Gc[:, :C] %= N_c[:C]
        Q1_Gc[:, :C] -= N_c[:C] // 2
        Q1_Gc[:, :C] %= N2_c[:C]
        Q2_G = Q1_Gc[:, 2] + N2_c[2] * (Q1_Gc[:, 1] + N2_c[1] * Q1_Gc[:, 0])
        G2_Q = np.empty(N2_c, int).ravel()
        G2_Q[:] = -1
        G2_Q[pw2.myindices] = np.arange(len(pw2.myindices))
        G2_G1 = G2_Q[Q2_G]

        if pw1.grid.comm.size == 1:
            self.G2_G1 = G2_G1
            self.G1 = None
        else:
            mask_G1 = (G2_G1 != -1)
            self.G2_G1 = G2_G1[mask_G1]
            self.G1 = np.arange(pw1.maxmysize)[mask_G1]

        self.pw1 = pw1
        self.pw2 = pw2

    def add_to1(self, a_G1, b_G2):
        """Do a += b * scale, where a is on pd1 and b on pd2."""
        scale = self.pd1.tmp_R.size / self.pd2.tmp_R.size

        if self.pd1.gd.comm.size == 1:
            a_G1 += b_G2[self.G2_G1] * scale
            return

        b_G1 = self.pd1.tmp_G
        b_G1[:] = 0.0
        b_G1[self.G1] = b_G2[self.G2_G1]
        self.pd1.gd.comm.sum(b_G1)
        ng1 = self.pd1.gd.comm.rank * self.pd1.maxmyng
        ng2 = ng1 + self.pd1.myng_q[0]
        a_G1 += b_G1[ng1:ng2] * scale

    def add_to2(self, a2, b1):
        """Do a += b * scale, where a is on pd2 and b on pd1."""
        myb = b1.data * (self.pw2.grid.shape[0] / self.pw1.grid.shape[0])
        if self.desc1.grid.comm.size == 1:
            a2.data[self.G2_G1] += myb
        else:
            1 / 0
'''



import pytest
import numpy as np
from gpaw.core.matrix import Matrix
from gpaw.mpi import world


@pytest.mark.skipif(world.size != 4, reason='size!=4')
def test_matrix_eigh():
    """Test eigenvalues for different BLACS layouts.

    See also #269.
    """
    N = 6
    x = 0.01

    A0 = Matrix(N, N, dist=(world, 1, 1), dtype=complex)

    if world.rank == 0:
        A0.data[:] = np.diag(np.arange(N) + 1)
        A0.data += np.random.uniform(-x, x, (N, N))
        A0.data += A0.data.conj().T
        B = Matrix(N, N, data=A0.data.copy())
        eigs0 = B.eigh(cc=True)
    else:
        eigs0 = np.empty(N)

    world.broadcast(eigs0, 0)

    A = Matrix(N, N, dist=(world, 2, 2, 2), dtype=complex)
    B0 = Matrix(N, N, dist=(world, 1, 1), dtype=complex)

    for dist in [(1, 2), (2, 1), (2, 2)]:
        A0.redist(A)
        eigs = A.eigh(cc=True, scalapack=(world, *dist, 2))
        assert eigs == pytest.approx(eigs0, abs=1e-13)
        print(world.rank, eigs)
        A.redist(B0)
        if world.rank == -1:
            assert abs(B0.data) == pytest.approx(abs(B.data), abs=1e-13)



import numpy as np
from ase.units import Bohr, Ha
from ase.build import molecule
from gpaw.cluster import Cluster
from gpaw import GPAW
from gpaw.external import static_polarizability


atoms = Cluster(molecule('H2O'))
atoms.minimal_box(3)
atoms.calc = GPAW(txt=None)

alpha_cc = static_polarizability(atoms)
print('Polarizability tensor (units Angstrom^3):')
print(alpha_cc * Bohr * Ha)

w, v = np.linalg.eig(alpha_cc)
print('Eigenvalues', w * Bohr * Ha)
print('Eigenvectors', v)
print('average polarizablity', w.sum() * Bohr * Ha / 3, 'Angstrom^3')



# This script helps to keep the documentation up-to-date.
# It checks that the values mentioned in .rst match
# with the actual output from the scripts.
#
# When updating the scripts, please update the .rst accordingly.

import numpy as np
from gpaw import GPAW


messages = []

calc = GPAW('unocc.gpw')
eig_n = calc.get_eigenvalues()

ref = 42
if not len(eig_n) == ref:
    messages.append(
        f'Number of bands is not {ref}. Check that discussion makes sense.')

ref = 8.2
if not np.isclose(eig_n[-2] - eig_n[11], ref, rtol=0, atol=0.1):
    messages.append(
        f'Update the energy difference of {ref}.')


data_ij = np.loadtxt('transitions_with_08.00eV.dat')
ref = 5.9
if not np.isclose(data_ij[0, 0], ref, rtol=0, atol=0.1):
    messages.append(
        f'Update the excitation energy of {ref}.')

if len(messages) > 0:
    messages.insert(0, 'Scripts have changed. Please update the .rst too:')
    message = '\n* '.join(messages)
    raise AssertionError(message)



from ase import Atoms
from ase.build import bulk
from ase.dft import monkhorst_pack
from ase.parallel import paropen
from gpaw import GPAW, FermiDirac
from gpaw import PW
from gpaw.hybrids.energy import non_self_consistent_energy as nsc_energy
import numpy as np

# Monkhorst-Pack grid shifted to be gamma centered
k = 8
kpts = monkhorst_pack([k, k, k])
kpts += [1. / (2 * k), 1. / (2 * k), 1. / (2 * k)]

cell = bulk('C', 'fcc', a=3.553).get_cell()
a = Atoms('C2', cell=cell, pbc=True,
          scaled_positions=((0, 0, 0), (0.25, 0.25, 0.25)))

calc = GPAW(mode=PW(600),
            xc='PBE',
            occupations=FermiDirac(width=0.01),
            convergence={'density': 1.e-6},
            kpts=kpts,
            parallel={'domain': 1},
            txt='diamond.ralda_01_pbe.txt')

a.calc = calc
E_pbe = a.get_potential_energy()

E_hf = nsc_energy(calc, 'ECC')

E_C = np.loadtxt('CO.ralda.PBE_HF_C.dat')

f = paropen('diamond.ralda.PBE_HF_diamond.dat', 'w')
print('PBE: ', E_pbe / 2 - E_C[0], file=f)
print('HF: ', E_hf / 2 - E_C[1], file=f)
f.close()

calc.diagonalize_full_hamiltonian()
calc.write('diamond.ralda.pbe_wfcs.gpw', mode='all')



from myqueue.workflow import run


def workflow():
    with run(script='C_ecut_k_conv_GW.py', cores=24, tmax='1d') as r:
        run(script='C_ecut_k_conv_plot_GW.py')
        run(script='C_ecut_extrap.py')

    with run(script='C_frequency_conv.py', tmax='30m'):
        with run(script='C_frequency_conv_plot.py'):
            run(script='C_equal_test.py', deps=[r])

    with run(script='MoS2_gs_GW.py', tmax='2h'):
        with run(script='MoS2_GWG.py', cores=8, tmax='20m'):
            run(script='MoS2_bs_plot.py')
            run(script='check_gw.py')



from __future__ import annotations

import importlib
from types import SimpleNamespace
from typing import Any, Union

import numpy as np
from ase import Atoms
from ase.calculators.calculator import kpts2sizeandoffsets
from ase.units import Bohr

from gpaw.core import UniformGrid
from gpaw.core.atom_arrays import (AtomArrays, AtomArraysLayout,
                                   AtomDistribution)
from gpaw.mixer import MixerWrapper, get_mixer_from_keywords
from gpaw.mpi import MPIComm, Parallelization, serial_comm, world
from gpaw.new import cached_property, prod
from gpaw.new.basis import create_basis
from gpaw.new.brillouin import BZPoints, MonkhorstPackKPoints
from gpaw.new.density import Density
from gpaw.new.input_parameters import InputParameters
from gpaw.new.scf import SCFLoop
from gpaw.new.smearing import OccupationNumberCalculator
from gpaw.new.symmetry import create_symmetries_object
from gpaw.new.xc import XCFunctional
from gpaw.setup import Setups
from gpaw.typing import DTypeLike, Array2D, ArrayLike1D, ArrayLike2D
from gpaw.utilities.gpts import get_number_of_grid_points


def builder(atoms: Atoms,
            params: dict[str, Any] | InputParameters) -> DFTComponentsBuilder:
    """Create DFT-components builder.

    * pw
    * lcao
    * fd
    * tb
    * atom
    """
    if isinstance(params, dict):
        params = InputParameters(params)

    mode = params.mode.copy()
    name = mode.pop('name')
    assert name in {'pw', 'lcao', 'fd', 'tb', 'atom'}
    mod = importlib.import_module(f'gpaw.new.{name}.builder')
    name = name.title() if name == 'atom' else name.upper()
    return getattr(mod, f'{name}DFTComponentsBuilder')(atoms, params, **mode)


class DFTComponentsBuilder:
    def __init__(self,
                 atoms: Atoms,
                 params: InputParameters):

        self.atoms = atoms.copy()
        self.mode = params.mode['name']
        self.params = params

        parallel = params.parallel
        world = parallel['world']

        self.check_cell(atoms.cell)

        self.initial_magmom_av, self.ncomponents = normalize_initial_magmoms(
            atoms, params.magmoms, params.spinpol or params.hund)

        self.soc = params.soc
        self.nspins = self.ncomponents % 3
        self.spin_degeneracy = self.ncomponents % 2 + 1

        self.xc = self.create_xc_functional()

        self.setups = Setups(atoms.numbers,
                             params.setups,
                             params.basis,
                             self.xc.setup_name,
                             world)

        if params.hund:
            c = params.charge / len(atoms)
            for a, setup in enumerate(self.setups):
                self.initial_magmom_av[a, 2] = setup.get_hunds_rule_moment(c)

        symmetries = create_symmetries_object(atoms,
                                              self.setups.id_a,
                                              self.initial_magmom_av,
                                              params.symmetry)
        assert not (self.ncomponents == 4 and len(symmetries) > 1)
        bz = create_kpts(params.kpts, atoms)
        self.ibz = symmetries.reduce(bz, strict=False)

        d = parallel.get('domain', None)
        k = parallel.get('kpt', None)
        b = parallel.get('band', None)
        self.communicators = create_communicators(world, len(self.ibz),
                                                  d, k, b)

        if self.mode == 'fd':
            pass  # filter = create_fourier_filter(grid)
            # setups = stups.filter(filter)

        self.nelectrons = self.setups.nvalence - params.charge

        self.nbands = calculate_number_of_bands(params.nbands,
                                                self.setups,
                                                params.charge,
                                                self.initial_magmom_av,
                                                self.mode == 'lcao')
        if self.ncomponents == 4:
            self.nbands *= 2

        self.dtype: DTypeLike
        if params.force_complex_dtype:
            self.dtype = complex
        elif self.ibz.bz.gamma_only:
            self.dtype = float
        else:
            self.dtype = complex

        self.grid, self.fine_grid = self.create_uniform_grids()

        self.fracpos_ac = self.atoms.get_scaled_positions()
        self.fracpos_ac %= 1
        self.fracpos_ac %= 1

    def create_uniform_grids(self):
        raise NotImplementedError

    def create_xc_functional(self):
        return XCFunctional(self.params.xc, self.ncomponents)

    def check_cell(self, cell):
        number_of_lattice_vectors = cell.rank
        if number_of_lattice_vectors < 3:
            raise ValueError(
                'GPAW requires 3 lattice vectors.  '
                f'Your system has {number_of_lattice_vectors}.')

    @cached_property
    def atomdist(self) -> AtomDistribution:
        return AtomDistribution(
            self.grid.ranks_from_fractional_positions(self.fracpos_ac),
            self.grid.comm)

    @cached_property
    def wf_desc(self):
        return self.create_wf_description()

    def __repr__(self):
        return f'{self.__class__.__name__}({self.atoms}, {self.params})'

    @cached_property
    def nct_R(self):
        out = self.grid.empty()
        nct_aX = self.get_pseudo_core_densities()
        nct_aX.to_uniform_grid(out=out,
                               scale=1.0 / (self.ncomponents % 3))
        return out

    def create_basis_set(self):
        return create_basis(self.ibz,
                            self.ncomponents % 3,
                            self.atoms.pbc,
                            self.grid,
                            self.setups,
                            self.dtype,
                            self.fracpos_ac,
                            self.communicators['w'],
                            self.communicators['k'])

    def density_from_superposition(self, basis_set):
        return Density.from_superposition(self.grid,
                                          self.nct_R,
                                          self.atomdist,
                                          self.setups,
                                          basis_set,
                                          self.initial_magmom_av,
                                          self.ncomponents,
                                          self.params.charge,
                                          self.params.hund)

    def create_occupation_number_calculator(self):
        return OccupationNumberCalculator(
            self.params.occupations,
            self.atoms.pbc,
            self.ibz,
            self.nbands,
            self.communicators,
            self.initial_magmom_av.sum(0),
            self.ncomponents,
            np.linalg.inv(self.atoms.cell.complete()).T)

    def create_scf_loop(self):
        hamiltonian = self.create_hamiltonian_operator()
        eigensolver = self.create_eigensolver(hamiltonian)

        mixer = MixerWrapper(
            get_mixer_from_keywords(self.atoms.pbc.any(),
                                    self.ncomponents, **self.params.mixer),
            self.ncomponents, self.grid._gd)

        occ_calc = self.create_occupation_number_calculator()

        return SCFLoop(hamiltonian, occ_calc,
                       eigensolver, mixer, self.communicators['w'],
                       self.params.convergence,
                       self.params.maxiter)

    def read_ibz_wave_functions(self, reader):
        raise NotImplementedError

    def create_potential_calculator(self):
        raise NotImplementedError

    def read_wavefunction_values(self, reader, ibzwfs):
        """ Read eigenvalues, occuptions and projections and fermi levels

        The values are read using reader and set as the appropriate properties
        of (the already instantiated) wavefunctions contained in ibzwfs
        """
        ha = reader.ha

        eig_skn = reader.wave_functions.eigenvalues
        occ_skn = reader.wave_functions.occupations
        P_sknI = reader.wave_functions.projections

        if self.params.force_complex_dtype:
            P_sknI = P_sknI.astype(complex)

        for wfs in ibzwfs:
            wfs._eig_n = eig_skn[wfs.spin, wfs.k] / ha
            wfs._occ_n = occ_skn[wfs.spin, wfs.k]
            layout = AtomArraysLayout([(setup.ni,) for setup in self.setups],
                                      dtype=self.dtype)
            if self.ncomponents < 4:
                wfs._P_ani = AtomArrays(layout,
                                        dims=(self.nbands,),
                                        data=P_sknI[wfs.spin, wfs.k])
            else:
                wfs._P_ani = AtomArrays(layout,
                                        dims=(self.nbands, 2),
                                        data=P_sknI[wfs.k])

        ibzwfs.fermi_levels = reader.wave_functions.fermi_levels / ha


def create_communicators(comm: MPIComm = None,
                         nibzkpts: int = 1,
                         domain: Union[int, tuple[int, int, int]] = None,
                         kpt: int = None,
                         band: int = None) -> dict[str, MPIComm]:
    parallelization = Parallelization(comm or world, nibzkpts)
    if domain is not None and not isinstance(domain, int):
        domain = prod(domain)
    parallelization.set(kpt=kpt,
                        domain=domain,
                        band=band)
    comms = parallelization.build_communicators()
    comms['w'] = comm
    return comms


def create_fourier_filter(grid):
    gamma = 1.6

    h = ((grid.icell**2).sum(1)**-0.5 / grid.size).max()

    def filter(rgd, rcut, f_r, l=0):
        gcut = np.pi / h - 2 / rcut / gamma
        ftmp = rgd.filter(f_r, rcut * gamma, gcut, l)
        f_r[:] = ftmp[:len(f_r)]

    return filter


def normalize_initial_magmoms(
        atoms: Atoms,
        magmoms: ArrayLike2D | ArrayLike1D | float | None = None,
        force_spinpol_calculation: bool = False) -> tuple[Array2D, int]:
    """Convert magnetic moments to (natoms, 3)-shaped array.

    Also return number of wave function components (1, 2 or 4).

    >>> h = Atoms('H', magmoms=[1])
    >>> normalize_initial_magmoms(h)
    (array([[0., 0., 1.]]), 2)
    >>> normalize_initial_magmoms(h, [[1, 0, 0]])
    (array([[1., 0., 0.]]), 4)
    """
    magmom_av = np.zeros((len(atoms), 3))
    ncomponents = 2

    if magmoms is None:
        magmom_av[:, 2] = atoms.get_initial_magnetic_moments()
    elif isinstance(magmoms, float):
        magmom_av[:, 2] = magmoms
    else:
        magmoms = np.asarray(magmoms)
        if magmoms.ndim == 1:
            magmom_av[:, 2] = magmoms
        else:
            magmom_av[:] = magmoms
            ncomponents = 4

    if (ncomponents == 2 and
        not force_spinpol_calculation and
        not magmom_av[:, 2].any()):
        ncomponents = 1

    return magmom_av, ncomponents


def create_kpts(kpts: dict[str, Any], atoms: Atoms) -> BZPoints:
    if 'points' in kpts:
        assert len(kpts) == 1, kpts
        return BZPoints(kpts['points'])
    size, offset = kpts2sizeandoffsets(**kpts, atoms=atoms)
    return MonkhorstPackKPoints(size, offset)


def calculate_number_of_bands(nbands: int | str | None,
                              setups: Setups,
                              charge: float,
                              initial_magmom_av: Array2D,
                              is_lcao: bool) -> int:
    nao = setups.nao
    nvalence = setups.nvalence - charge
    M = np.linalg.norm(initial_magmom_av.sum(0))

    orbital_free = any(setup.orbital_free for setup in setups)
    if orbital_free:
        return 1

    if isinstance(nbands, str):
        if nbands == 'nao':
            N = nao
        elif nbands[-1] == '%':
            cfgbands = (nvalence + M) / 2
            N = int(np.ceil(float(nbands[:-1]) / 100 * cfgbands))
        else:
            raise ValueError('Integer expected: Only use a string '
                             'if giving a percentage of occupied bands')
    elif nbands is None:
        # Number of bound partial waves:
        nbandsmax = sum(setup.get_default_nbands()
                        for setup in setups)
        N = int(np.ceil((1.2 * (nvalence + M) / 2))) + 4
        if N > nbandsmax:
            N = nbandsmax
        if is_lcao and N > nao:
            N = nao
    elif nbands <= 0:
        N = max(1, int(nvalence + M + 0.5) // 2 + (-nbands))
    else:
        N = nbands

    if N > nao and is_lcao:
        raise ValueError('Too many bands for LCAO calculation: '
                         f'{nbands}%d bands and only {nao} atomic orbitals!')

    if nvalence < 0:
        raise ValueError(
            f'Charge {charge} is not possible - not enough valence electrons')

    if nvalence > 2 * N:
        raise ValueError(
            f'Too few bands!  Electrons: {nvalence}, bands: {nbands}')

    return N


def create_uniform_grid(mode: str,
                        gpts,
                        cell,
                        pbc,
                        symmetry,
                        h: float = None,
                        interpolation: str = None,
                        ecut: float = None,
                        comm: MPIComm = serial_comm) -> UniformGrid:
    """Create grid in a backwards compatible way."""
    cell = cell / Bohr
    if h is not None:
        h /= Bohr

    realspace = (mode != 'pw' and interpolation != 'fft')
    if not realspace:
        pbc = (True, True, True)

    if gpts is not None:
        size = gpts
    else:
        modeobj = SimpleNamespace(name=mode, ecut=ecut)
        size = get_number_of_grid_points(cell, h, modeobj, realspace,
                                         symmetry.symmetry)
    return UniformGrid(cell=cell, pbc=pbc, size=size, comm=comm)



import pytest

from gpaw import GPAW, LCAO
from gpaw.directmin.derivatives import Derivatives
from ase import Atoms


def test_gradient_numerically_lcao(in_tmp_dir):
    """
    test exponential transformation
    direct minimization method for KS-DFT in LCAO
    :param in_tmp_dir:
    :return:
    """

    # Water molecule:
    atoms = Atoms('H3', positions=[(0, 0, 0),
                                   (0.59, 0, 0),
                                   (1.1, 0, 0)])
    atoms.center(vacuum=2.0)
    atoms.set_pbc(True)
    calc = GPAW(mode=LCAO(force_complex_dtype=True),
                basis='sz(dzp)',
                h=0.3,
                spinpol=False,
                convergence={'eigenstates': 10.0,
                             'density': 10.0,
                             'energy': 10.0},
                occupations={'name': 'fixed-uniform'},
                eigensolver={'name': 'etdm',
                             'matrix_exp': 'egdecomp'},
                mixer={'backend': 'no-mixing'},
                nbands='nao',
                symmetry='off',
                txt=None
                )
    atoms.calc = calc
   
    params = [{'name': 'etdm',
               'representation': 'full',
               'matrix_exp': 'egdecomp'},
              {'name': 'etdm',
               'representation': 'full',
               'matrix_exp': 'pade-approx'},
              {'name': 'etdm',
               'representation': 'sparse',
               'matrix_exp': 'egdecomp'},
              {'name': 'etdm',
               'representation': 'sparse',
               'matrix_exp': 'pade-approx'},
              {'name': 'etdm',
               'representation': 'u-invar',
               'matrix_exp': 'egdecomp'},
              {'name': 'etdm',
               'representation': 'u-invar',
               'matrix_exp': 'egdecomp-u-invar'}
              ]

    for eigsolver in params:
        print('IN PROGRESS: ', eigsolver)

        calc.set(eigensolver=eigsolver)
        if eigsolver['representation'] == 'u-invar':
            with pytest.warns(UserWarning,
                              match="Use representation == 'sparse'"):
                atoms.get_potential_energy()
        else:
            atoms.get_potential_energy()
        ham = calc.hamiltonian
        wfs = calc.wfs
        dens = calc.density

        if eigsolver['matrix_exp'] == 'egdecomp':
            update_c_ref = False
        else:
            update_c_ref = True

        numder = Derivatives(wfs.eigensolver, wfs, random_amat=True,
                             update_c_ref=update_c_ref)

        g_a = numder.get_analytical_derivatives(wfs.eigensolver, ham, wfs,
                                                dens)
        g_n = numder.get_numerical_derivatives(wfs.eigensolver, ham, wfs, dens)

        for x, y in zip(g_a[0], g_n[0]):
            assert x.real == pytest.approx(y.real, abs=1.0e-2)
            assert x.imag == pytest.approx(y.imag, abs=1.0e-2)



#!/usr/bin/env python
# flake8: noqa
from optparse import OptionParser

code_choices = ['gpaw', 'dacapo']

parser = OptionParser(usage='%prog [options] package.\nExample of call:\n'+
                      'python %prog\n',
                      version='%prog 0.1')
parser.add_option('--code', dest="code", type="choice",
                  default=code_choices[0],
                  choices=code_choices,
                  help='code: which code to use.')
parser.add_option("--runs", dest="runs",
                  default=7,
                  help='use that many runs to calculate the average.')
parser.add_option('-v', '--verbose', action='store_true',
                  default=False,
                  help='verbose mode.')


opt, args = parser.parse_args()

from os import remove
from os.path import exists

try:
    import numpy as np
except ImportError:
    raise SystemExit('numpy is not installed!')

try:
    import gpaw
except ImportError:
    raise SystemExit('gpaw is not installed!')

from gpaw.utilities.tools import gridspacing2cutoff

try:
    import ase
except ImportError:
    raise SystemExit('ase is not installed!')

from ase import Atoms, Atom

import time

a = 4.00
d = a / 2**0.5
z = 1.1
b = 1.5

def memory_bandwidth(code='gpaw', runs=7):

    slab = Atoms([Atom('Al', (0, 0, 0)),
                        Atom('Al', (a, 0, 0)),
                        Atom('Al', (a/2, d/2, -d/2)),
                        Atom('Al', (3*a/2, d/2, -d/2)),
                        Atom('Al', (0, 0, -d)),
                        Atom('Al', (a, 0, -d)),
                        Atom('Al', (a/2, d/2, -3*d/2)),
                        Atom('Al', (3*a/2, d/2, -3*d/2)),
                        Atom('Al', (0, 0, -2*d)),
                        Atom('Al', (a, 0, -2*d)),
                        Atom('H', (a/2-b/2, 0, z)),
                        Atom('H', (a/2+b/2, 0, z))],
                       cell=(2*a, d, 5*d), pbc=(1, 1, 1))

    h = 0.15
    nbands = 28
    kpts = (2, 6, 1)

    parameters = {}

    if code == 'gpaw':
        from gpaw import Calculator
        from gpaw.mpi import rank
        parameters['convergence'] = {'eigenstates': 1e-5}
        parameters['h'] = h
    elif code == 'dacapo':
        from ase.calculators.dacapo import Dacapo as Calculator
        parameters['planewavecutoff'] = gridspacing2cutoff(h)
        parameters['densitycutoff'] = parameters['planewavecutoff']*1.5
        rank = 0

    t = 0.0
    t_runs = []
    for n in range(runs):
        t0 = time.time()
        for i in range(1):
            calc = Calculator(
                nbands=nbands,
                kpts=kpts,
                **parameters)
            slab.calc = calc
            e = slab.get_potential_energy()
            del calc
            if exists('out.nc'): remove('out.nc')
        t1 = time.time()
        t = t + t1 - t0
        t_runs.append(t1 - t0)
        print('Run: ', n, ' energy ', e, ' rank: ', str(rank), ' time: ', time.time() - t0)
    if rank == 0:
        print('Rank '+str(rank)+': time [sec]: avg '+str(round(np.average(t_runs),1))+', stddev '+str(round(np.std(t_runs),1))+', min '+str(round(min(t_runs),1))+', max '+str(round(max(t_runs),1)))


if __name__ == '__main__':


    code = opt.code
    assert code in code_choices, code+' not in '+str(code_choices)

    if code == 'dacapo':
        try:
            import ASE
        except ImportError:
            raise SystemExit('ASE (2) is not installed!')

    runs = int(opt.runs)
    assert runs >= 1, runs+' must be >= 1'

    memory_bandwidth(code=code, runs=runs)



# web-page: bandstructure.png
"""Band structure tutorial

Calculate the band structure of Si along high symmetry directions
Brillouin zone
"""
# P1
from ase.build import bulk
from gpaw import GPAW, PW, FermiDirac

# Perform standard ground state calculation (with plane wave basis)
si = bulk('Si', 'diamond', 5.43)
calc = GPAW(mode=PW(200),
            xc='PBE',
            kpts=(8, 8, 8),
            random=True,  # random guess (needed if many empty bands required)
            occupations=FermiDirac(0.01),
            txt='Si_gs.txt')
si.calc = calc
si.get_potential_energy()
ef = calc.get_fermi_level()
calc.write('Si_gs.gpw')
# P2
# Restart from ground state and fix potential:
calc = GPAW('Si_gs.gpw').fixed_density(
    nbands=16,
    symmetry='off',
    kpts={'path': 'GXWKL', 'npoints': 60},
    convergence={'bands': 8})

# P3
bs = calc.band_structure()
bs.plot(filename='bandstructure.png', show=True, emax=10.0)



import numpy as np
from gpaw.test import equal

results = np.loadtxt('diam_kern.ralda_kernel_comparison.dat')

ralda_wave = results[0]
raldac = results[1]
JGMs = results[2]
CP_dyn = results[3]
range_RPA = results[4]
RPA = results[5]

equal(ralda_wave, -9.08, 0.01)
equal(raldac, -9.01, 0.01)
equal(JGMs, -9.07, 0.01)
equal(CP_dyn, -8.55, 0.01)
equal(range_RPA, -13.84, 0.01)
equal(RPA, -11.17, 0.01)



import numpy as np
from ase import Atoms
# from ase.units import Ha
from gpaw import GPAW, PW
# from gpaw.hybrids import HybridXC
# from gpaw.hybrids.eigenvalues import non_self_consistent_eigenvalues
from gpaw.hybrids.energy import non_self_consistent_energy
from gpaw.hybrids.eigenvalues import non_self_consistent_eigenvalues
from gpaw.xc.exx import EXX
from gpaw.mpi import world, serial_comm


def test(kpts, setup, spinpol, symmetry):
    a = Atoms('HLi',
              cell=(3, 3, 3),
              pbc=1,
              positions=[[1.4, 0, 1.05], [1.6, 0, 2]]
              # positions=[[0, 0, 1.25], [0, 0, 2]]
              # positions=[[0, 0, 0], [0, 0, 0.75]]
              )
    a.calc = GPAW(mode=PW(100, force_complex_dtype=True),
                  setups=setup,
                  kpts=kpts,
                  spinpol=spinpol,
                  # nbands=1,
                  symmetry=symmetry,
                  # parallel={'band': 1, 'kpt': 1},
                  txt=None,
                  xc='PBE')
    a.get_potential_energy()
    return a


def check(atoms, xc, i):
    # xc1 = HybridXC(xc)
    c = atoms.calc
    # xc1.initialize(c.density, c.hamiltonian, c.wfs, c.occupations)
    # xc1.set_positions(c.spos_ac)
    e = non_self_consistent_energy(c, xc)
    c.write('c.gpw', 'all')
    eps = non_self_consistent_eigenvalues(c, xc, 0, 2, snapshot=f'{i}.json')
    # eps = non_self_consistent_eigenvalues('c.gpw', xc, 0, 2,
    #                                      snapshot=f'{i}.json')
    # xc1.calculate_eigenvalues0(0, 2, None)
    # e1, v1, v2 = non_self_consistent_eigenvalues(c, xc, 0, 2, None,
    #                                             f'{i}.txt')
    if world.size > 1:
        c.write('tmp.gpw', 'all')
        c = GPAW('tmp.gpw', communicator=serial_comm, txt=None)

    xc2 = EXX(c.gs_adapter(), xc=xc, bands=(0, 2), txt=None)
    xc2.calculate()
    # e0 = xc2.get_exx_energy()
    et0 = xc2.get_total_energy()
    eps0 = xc2.get_eigenvalue_contributions()
    assert np.allclose(eps0, eps[2]), (eps0, eps)
    # assert np.allclose(v2, xc1.e_skn * Ha), (v2, xc1.e_skn * Ha, eps0)
    # assert np.allclose(eps0, xc1.e_skn * Ha)
    # print(e0, e)
    # assert np.allclose(e0, e[-3:].sum())
    # print(et0, e);asdfg
    assert np.allclose(et0, e.sum()), (et0, e)
    # ecv, evv, v_skn = xc1.test()
    # assert np.allclose(e0, ecv + evv)
    # assert np.allclose(v_skn, eps0)


def main():
    i = 0
    for spinpol in [False, True]:
        for setup in [
            # 'ae',
            'paw']:
            for symmetry in ['off', {}]:
                for kpts in [(1, 1, 1),
                             (1, 1, 2),
                             (1, 1, 3),
                             (1, 1, 4),
                             (2, 2, 1),
                             [(0, 0, 0.5)],
                             [(0, 0, 0), (0, 0, 0.5)]]:
                    atoms = test(kpts, setup, spinpol, symmetry)
                    for xc in ['EXX',
                               'PBE0', 'HSE06']:
                        print(i, spinpol, setup, symmetry, kpts, xc,
                              len(atoms.calc.wfs.kpt_u))
                        check(atoms, xc, i)
                        i += 1


main()



from gpaw.tddft.spectrum import rotatory_strength_spectrum

rotatory_strength_spectrum(['mm-x.dat', 'mm-y.dat', 'mm-z.dat'],
                           'rot_spec.dat',
                           folding='Gauss', width=0.2,
                           e_min=0.0, e_max=10.0, delta_e=0.01)



import numpy as np
import pytest
from ase import Atoms
from gpaw import GPAW, restart
from gpaw.mom import prepare_mom_calculation
from gpaw.directmin.etdm import ETDM
from gpaw.directmin.tools import excite


@pytest.mark.mom
def test_mom_directopt_lcao_forces(in_tmp_dir):
    L = 4.0
    d = 1.13
    delta = 0.01

    atoms = Atoms('CO',
                  [[0.5 * L, 0.5 * L, 0.5 * L - 0.5 * d],
                   [0.5 * L, 0.5 * L, 0.5 * L + 0.5 * d]])
    atoms.set_cell([L, L, L])
    atoms.set_pbc(True)

    calc = GPAW(mode='lcao',
                basis='dzp',
                h=0.22,
                xc='PBE',
                spinpol=True,
                symmetry='off',
                occupations={'name': 'fixed-uniform'},
                eigensolver={'name': 'etdm',
                             'linesearch_algo': 'max-step'},
                mixer={'backend': 'no-mixing'},
                nbands='nao',
                convergence={'density': 1.0e-4,
                             'eigenstates': 4.0e-8})
    atoms.calc = calc
    atoms.get_potential_energy()

    f_sn = excite(calc, 0, 0, spin=(0, 0))

    calc.set(eigensolver=ETDM(searchdir_algo={'name': 'l-sr1p'},
                              linesearch_algo={'name': 'max-step'},
                              representation='u-invar',
                              matrix_exp='egdecomp-u-invar',
                              need_init_orbs=False))
    prepare_mom_calculation(calc, atoms, f_sn)
    F = atoms.get_forces()

    # Test overlaps
    calc.wfs.occupations.initialize_reference_orbitals()
    for kpt in calc.wfs.kpt_u:
        f_n = calc.get_occupation_numbers(spin=kpt.s)
        unoccupied = [True for _ in range(len(f_n))]
        P = calc.wfs.occupations.calculate_weights(kpt, 1.0, unoccupied)
        assert (np.allclose(P, f_n))

    calc.write('co.gpw', mode='all')

    # Exercise fixed occupations and no update of numbers in OccupationsMOM
    atoms, calc = restart('co.gpw', txt='-')
    e0 = atoms.get_potential_energy()
    for i in [True, False]:
        prepare_mom_calculation(calc, atoms, f_sn,
                                use_fixed_occupations=i,
                                update_numbers=i)
        e1 = atoms.get_potential_energy()
        for spin in range(calc.get_number_of_spins()):
            if i:
                f_n = calc.get_occupation_numbers(spin=spin)
                assert (np.allclose(f_sn[spin], f_n))
            assert (np.allclose(f_sn[spin],
                                calc.wfs.occupations.numbers[spin]))
        assert e0 == pytest.approx(e1, abs=1e-2)

    E = []
    for i in [-1, 1]:
        atoms, calc = restart('co.gpw', txt='-')
        p = atoms.positions.copy()
        p[0, 2] -= delta / 2. * i
        p[1, 2] += delta / 2. * i
        atoms.set_positions(p)
        E.append(atoms.get_potential_energy())

    f = np.sqrt(((F[1, :] - F[0, :])**2).sum()) * 0.5
    fnum = (E[0] - E[1]) / (2. * delta)     # central difference

    print(fnum)
    assert fnum == pytest.approx(12.407162321236331, abs=0.01)
    assert f == pytest.approx(fnum, abs=0.1)



import numpy as np
import pytest

from ase.phonons import Phonons
from ase.utils.filecache import MultiFileJSONCache
from gpaw import GPAW
from gpaw.raman.elph import EPC


class FakeEPC(EPC):
    """Fake ElectronPhononCoupling class to overwrite loading routine."""
    def __init__(self, fake_basis, natom, **params):
        EPC.__init__(self, **params)
        # self.gx = np.random.random(size=[3, 1, 1, 1, 4, 4])
        gx = np.arange(3 * natom * 4 * 4).reshape([3 * natom, 1, 1, 1, 4, 4])

        # Fake supercell cache
        self.supercell_cache = MultiFileJSONCache('supercell')
        with self.supercell_cache.lock('basis') as handle:
            if handle is not None:
                handle.save(fake_basis)
        for x in range(3 * natom):
            with self.supercell_cache.lock(str(x)) as handle:
                if handle is None:
                    continue
                handle.save(gx[x])


class FakePh(Phonons):
    """Fake Phonons object class to overwrite reading routine."""
    def read(self):
        natoms = len(self.indices)
        m_a = self.atoms.get_masses()
        self.m_inv_x = np.repeat(m_a[self.indices]**-0.5, 3)
        if self.D_N is None:
            # self.D_N = np.random.random([3 * natoms, 3 * natoms])
            self.D_N = np.ones([3 * natoms, 3 * natoms])


@pytest.mark.serial
def test_elph_matrix(gpw_files, tmp_path_factory):
    """Test of elph_matrix function as well as load_gx_as_needed feature."""
    calc = GPAW(gpw_files['bcc_li_lcao_wfs'])
    atoms = calc.atoms
    # Initialize calculator if necessary
    if not hasattr(calc.wfs, 'C_nM'):
        calc.wfs.set_positions
        calc.initialize_positions(atoms)

    for kpt in calc.wfs.kpt_u:
        print(kpt.eps_n)

    # create phonon object
    phonon = FakePh(atoms)
    # create an electron-phonon object
    elph = FakeEPC(fake_basis={'M_a': [0], 'nao_a': [4]}, natom=1,
                   atoms=atoms)
    elph.calc_lcao = calc  # set calc directly to circumvent some checks

    g_sqklnn = elph.get_elph_matrix(calc, phonon, savetofile=False)
    assert g_sqklnn.shape == (1, 1, 4, 3, 4, 4)

    # quick check of phonon object
    phonon.read()
    frequencies, modes = phonon.band_structure([[0., 0., 0.]], modes=True)
    modes = modes.reshape(3, 3).real
    assert modes[2] == pytest.approx([0.21915917, 0.21915917, 0.21915917])

    # this should always be the same, as long as CnM doesn't change
    # there are plenty of generate bands. will lead of different results on
    # different machines for some bands
    tmp = g_sqklnn[0, 0, :, 2].real
    print(tmp)
    # bands 1-3 degenerate for this kpoint, don't use
    assert tmp[0, 0, 0] == pytest.approx(8.57864400e+01)
    # bands 2-3 degenerate for this kpoint, don't use
    assert tmp[1, 0, 0] == pytest.approx(2.49398305e+02)
    # absolute value because of random phase
    assert abs(tmp[1, 0, 1]) == pytest.approx(4.15645630e+02)
    assert tmp[1, 1, 1] == pytest.approx(1.41967585e+04)



import matplotlib.pyplot as plt
from ase import Atoms
from ase.units import Bohr
from gpaw.new.ase_interface import GPAW

plt.rcParams['font.size'] = 12

hli = Atoms('HLi', positions=[[0, 0, 0], [0, 0, 1.6]])
hli.center(vacuum=2.5)
hli.calc = GPAW(txt='hli.txt', mode='fd')
hli.get_potential_energy()

for n in range(2):
    ae = hli.calc.calculation.state.ibzwfs.get_all_electron_wave_function(
        n, grid_spacing=0.05)
    norm_squared = ae.norm2()
    print('Norm:', norm_squared)
    assert abs(norm_squared - 1) < 1e-2

    # Interpolated AE wfs:
    i0, i1 = ae.desc.size[:2] // 2
    x, y = ae.xy(i0, i1, ...)
    x *= Bohr
    y *= Bohr**-1.5
    plt.plot(x, y, '-', color=f'C{n}', label=rf'$\psi_{n}$')

    # Raw PS wfs:
    ps = hli.calc.calculation.state.ibzwfs.wfs_qs[0][0].psit_nX[n]
    i0, i1 = ps.desc.size[:2] // 2 - 1
    x, y = ps.xy(i0, i1, ...)
    x *= Bohr
    y *= Bohr**-1.5
    plt.plot(x, y, 'o', color=f'C{n}', label=rf'$\tilde\psi_{n}$')

plt.hlines(0, xmin=x[0], xmax=x[-1])
plt.xlabel(r'z [$\rm \AA$]')
plt.ylabel(r'wave functions [$\rm \AA^{-3/2}$]')
plt.legend()
plt.tight_layout()
plt.savefig('hli-wfs.png')
hli.calc.write('hli.gpw')
plt.close()



"""Test CLI-dipole script."""
import pytest
from gpaw.utilities.dipole import main


@pytest.mark.serial
@pytest.mark.parametrize(
    'mode',
    ['pw', 'lcao', 'fd'])
def test_dipole(gpw_files, mode, capsys):
    args = [str(gpw_files[f'h2_{mode}_wfs'])]
    main(args)
    out = capsys.readouterr().out
    lines = [line.split() for line in out.splitlines()]
    assert abs(float(lines[6][2])) == pytest.approx(0.55, abs=0.05)



import matplotlib.pyplot as plt
from ase.units import Bohr
from gpaw.new.ase_interface import GPAW

calc = GPAW('hli.gpw', txt=None)
elpot = calc.calculation.electrostatic_potential()

# Avarage PS potentials:
vh, vli = elpot.atomic_potentials()
zh, zli = calc.atoms.positions[:, 2]
rh, rli = [max(setup.rcut_j) * Bohr for setup in elpot.setups]
plt.plot([zh - rh, zh + rh], [vh, vh], label=r'$\tilde v$(H)')
plt.plot([zli - rli, zli + rli], [vli, vli], label=r'$\tilde v$(Li)')

# Interpolated PS and AE potentials:
ps = elpot.pseudo_potential(grid_spacing=0.025)
ae = elpot.all_electron_potential(grid_spacing=0.025)
i = ps.data.shape[0] // 2

x, y = ps.xy(i, i, ...)
plt.plot(x, y, '-', label=r'$\tilde v$')

x, y = ae.xy(i, i, ...)
plt.plot(x, y, '-', label=r'$v$')

plt.plot(x, 0 * x, 'k')

plt.xlabel('z [Ang]')
plt.ylabel('potential [eV]')
plt.ylim(bottom=-100, top=10)
plt.legend()
plt.savefig('hli-pot.png')



"""Test functionality to compute Fourier Transforms with PAW corrections"""

# General modules
import numpy as np

# Script modules
from ase.build import bulk

from gpaw import GPAW, PW, FermiDirac
from gpaw.response.mft import PlaneWaveBxc
from gpaw.response.susceptibility import get_pw_coordinates
from gpaw.test.response.test_site_kernels import get_PWDescriptor


# ---------- Actual tests ---------- #


def test_Fe_bxc(in_tmp_dir):
    """Test the symmetry relation

    (B^xc_G)^* = B^xc_-G

    for a real life system with d-electrons (bcc-Fe)."""
    # ---------- Inputs ---------- #

    # Part 1: Ground state calculation
    xc = 'LDA'
    kpts = 4
    nbands = 6
    pw = 200
    occw = 0.01
    conv = {'density': 1e-8,
            'forces': 1e-8,
            'bands': nbands}
    a = 2.867
    mm = 2.21

    # Part 2: Bxc calculation
    ecut = 100

    # Part 3: Check symmetry relation

    # ---------- Script ---------- #

    # Part 1: Ground state calculation

    atoms = bulk('Fe', 'bcc', a=a)
    atoms.set_initial_magnetic_moments([mm])
    atoms.center()

    calc = GPAW(xc=xc,
                mode=PW(pw),
                kpts={'size': (kpts, kpts, kpts), 'gamma': True},
                nbands=nbands + 4,
                occupations=FermiDirac(occw),
                parallel={'domain': 1},
                spinpol=True,
                convergence=conv
                )

    atoms.calc = calc
    atoms.get_potential_energy()

    # Part 2: Bxc calculation
    Bxc_calc = PlaneWaveBxc(calc)
    pd0 = get_PWDescriptor(atoms, calc, [0., 0., 0.],
                           ecut=ecut,
                           gammacentered=True)

    Bxc_G = Bxc_calc(pd0)

    # Part 3: Check symmetry relation
    G1_G, G2_G = get_inversion_pairs(pd0)

    assert np.allclose(np.conj(Bxc_G[G1_G]), Bxc_G[G2_G])


# ---------- Test functionality ---------- #


def get_inversion_pairs(pd0):
    """Get all pairs of G-indices which correspond to inverted reciprocal
    lattice vectors G and -G."""
    G_Gc = get_pw_coordinates(pd0)

    G1_G = []
    G2_G = []
    paired_indices = []
    for G1, G1_c in enumerate(G_Gc):
        if G1 in paired_indices:
            continue  # Already paired

        for G2, G2_c in enumerate(G_Gc):
            if np.all(G2_c == -G1_c):
                G1_G.append(G1)
                G2_G.append(G2)
                paired_indices += [G1, G2]
                break

    assert len(np.unique(paired_indices)) == len(G_Gc)

    return G1_G, G2_G



from ase import Atoms
from ase.build import molecule
from ase.parallel import paropen
from gpaw import GPAW, Mixer, MixerDif
from gpaw.utilities.tools import split_formula

cell = [14.4, 14.4, 14.4]
data = paropen('data.txt', 'a')

# Reference from J. Chem. Phys. Vol 120 No. 15, 15 April 2004, page 6898
tpss_de = [
    ('H2', 112.9),
    ('LiH', 59.1),
    ('OH', 106.8),
    ('HF', 139.1),
    ('Li2', 22.5),
    ('LiF', 135.7),
    ('Be2', 8.1),
    ('CO', 254.2),
    ('N2', 227.7),
    ('O2', 126.9),
    ('F2', 46.4),
    ('P2', 116.1),
    ('Cl2', 60.8)]

exp_bonds_dE = [
    ('H2', 0.741, 109.5),
    ('LiH', 1.595, 57.8),
    ('OH', 0.970, 106.4),
    ('HF', 0.917, 140.8),
    ('Li2', 2.673, 24.4),
    ('LiF', 1.564, 138.9),
    ('Be2', 2.440, 3.0),
    ('CO', 1.128, 259.3),
    ('N2', 1.098, 228.5),
    ('O2', 1.208, 120.5),
    ('F2', 1.412, 38.5),
    ('P2', 1.893, 117.3),
    ('Cl2', 1.988, 58.0)]

systems = [a[0] for a in tpss_de]
ref = [a[1] for a in tpss_de]

# Add atoms
for formula in systems:
    temp = split_formula(formula)
    for atom in temp:
        if atom not in systems:
            systems.append(atom)
energies = {}

# Calculate energies
i = 0
for formula in systems:
    if formula == 'Be2':
        loa = Atoms('Be2', [(0, 0, 0), (0, 0, 2.0212)])
    else:
        loa = molecule(formula)
    loa.set_cell(cell)
    loa.center()
    width = 0.0
    calc = GPAW(h=.18,
                nbands=-5,
                maxiter=333,
                xc='PBE',
                txt=formula + '.txt')
    if len(loa) == 1:
        calc.set(hund=True)
        # calc.set(fixmom=True)
        calc.set(mixer=MixerDif())
        calc.set(eigensolver='cg')
    else:
        calc.set(mixer=Mixer())
        pos = loa.get_positions()
        pos[1, :] = pos[0, :] + [exp_bonds_dE[i][1], 0.0, 0.0]
        loa.set_positions(pos)
        loa.center()
    loa.calc = calc
    try:
        energy = loa.get_potential_energy()
        difft = calc.get_xc_difference('TPSS')
        diffr = calc.get_xc_difference('revTPSS')
        diffm = calc.get_xc_difference('M06-L')
        energies[formula] = (energy, energy + difft, energy + diffr,
                             energy + diffm)
    except Exception:
        print(formula, 'Error', file=data)
    else:
        print(formula, energy, energy + difft, energy + diffr, energy + diffm,
              file=data)
        data.flush()
    i += 1

# Calculate atomization energies
ii = 0
file = paropen('atom_en.dat', 'a')
print('# formula \t PBE \t TPSS \t revTPSS \t M06-L \t Exp', file=file)
for formula in systems[:13]:
    try:
        atoms_formula = split_formula(formula)
        de_tpss = -1.0 * energies[formula][1]
        de_revtpss = -1.0 * energies[formula][2]
        de_m06l = -1.0 * energies[formula][3]
        de_pbe = -1.0 * energies[formula][0]
        for atom_formula in atoms_formula:
            de_tpss += energies[atom_formula][1]
            de_revtpss += energies[atom_formula][2]
            de_m06l += energies[atom_formula][3]
            de_pbe += energies[atom_formula][0]
    except Exception:
        print(formula, 'Error', file=file)
    else:
        de_tpss *= 627.5 / 27.211
        de_revtpss *= 627.5 / 27.211
        de_m06l *= 627.5 / 27.211
        de_pbe *= 627.5 / 27.211
        out = ('%s\t%.1f \t%.1f \t%.1f \t%.1f \t%.1f' %
               (formula, de_pbe, de_tpss, de_revtpss, de_m06l,
                exp_bonds_dE[ii][2]))
        print(out, file=file)
        file.flush()
    ii += 1



import pytest
from ase import Atoms
from ase.units import Ha

from gpaw.new.ase_interface import GPAW


@pytest.mark.serial
@pytest.mark.xfail
def test_h_exx_lcao():
    atoms = Atoms('H', magmoms=[1])
    atoms.center(vacuum=2.5)
    atoms.calc = GPAW(mode='lcao',
                      xc='EXX')
    atoms.get_potential_energy()
    eig = atoms.calc.get_eigenvalues(spin=0)[0]
    assert eig == pytest.approx(-0.5 * Ha, abs=0.05)



# web-page: overview.png

import numpy as np
from ase.io.cube import read_cube
from matplotlib import pyplot
from ase.data import covalent_radii as radii
from ase.data.colors import jmol_colors
from matplotlib.patches import Circle


fig, ax = pyplot.subplots(figsize=(6., 1.8))
fig.subplots_adjust(left=0.08, right=0.99, bottom=0.23, top=0.99)

###########################
# Plot solvent.
###########################

with open('sjm_traces4.4V-cube.out/cavity.cube') as f:
    data = read_cube(f)

atoms = data['atoms']
cube = data['data']
flat = np.mean(cube, axis=0)

ax.imshow(flat, cmap='Blues', interpolation='spline16',
          origin='lower', vmax=1.5,
          extent=(0., atoms.cell[2, 2], 0., atoms.cell[1, 1]))

###########################
# Add atoms as circles.
###########################

cell = atoms.cell
atomslist = [atom for atom in atoms]
atomslist = sorted(atomslist, key=lambda atom: atom.x)
atomslist.reverse()

# Add the atoms to the plot as circles.
for atom in atomslist:
    color = jmol_colors[atom.number]
    radius = radii[atom.number]
    circle = Circle((atom.z, atom.y), radius, facecolor=color,
                    edgecolor='k', linewidth=1)
    ax.add_patch(circle)

xlim = ax.get_xlim()
ylim = ax.get_ylim()

###########################
# Add the jellium region.
###########################

with open('sjm_traces4.4V.out/background_charge.txt', 'r') as f:
    lines = f.read().splitlines()

zs = []
jelliums = []
for line in lines:
    z, jellium = line.split()
    zs.append(float(z))
    jelliums.append(float(jellium))

jelliums = np.array(jelliums) * atoms.cell[1][1]

where = [True if jellium != 0 else False for jellium in jelliums]
ax.fill_between(zs, jelliums, where=where, hatch='///', ec='0.2')

###########################
# Touch up.
###########################

ax.text(24., max(jelliums) / 2., 'jellium', ha='center', va='center')
ax.text(17.8, max(jelliums) / 2., 'solvent', ha='center',
        va='center', rotation=90.)
ax.set_xlim(xlim)
ax.set_ylim(ylim)
ax.set_xlabel(r'$z$, $\mathrm{\AA}$')
ax.set_ylabel(r'$y$, $\mathrm{\AA}$')

fig.savefig('overview.png')



from pathlib import Path

from ase import Atoms
from gpaw import GPAW, PW
from gpaw.mpi import size
from gpaw.response.g0w0 import G0W0

c = 2.53574055
atoms = Atoms('ZrO2',
              scaled_positions=[(0, 0, 0),
                                (0.25, 0.25, 0.25),
                                (0.75, 0.75, 0.75)],
              cell=[(0, c, c), (c, 0, c), (c, c, 0)],
              pbc=True)


def gs(k: int, ecut1: float) -> None:
    atoms.calc = GPAW(mode=PW(ecut1),
                      kpts={'size': (k, k, k), 'gamma': True},
                      xc='PBEsol',
                      txt='gs.txt')
    atoms.get_potential_energy()
    atoms.calc.diagonalize_full_hamiltonian()
    atoms.calc.write('gs.gpw', mode='all')


def gw(ecut2: float) -> dict:
    gw = G0W0(calc='gs.gpw',
              relbands=(-3, 3),
              ecut=ecut2,
              frequencies={'type': 'nonlinear',
                           'domega0': 0.025,
                           'omega2': 10.0},
              eta=0.1,
              filename='gw',
              nblocks=size // 2)
    result = gw.calculate()
    return result


def main():
    if not Path('gs.gpw').is_file():
        gs(k=4, ecut1=600)
    gw(ecut2=350)


if __name__ == '__main__':
    main()



from ase.units import Ha
import os
import gpaw.mpi as mpi
import numpy as np
from gpaw.xc.fxc import KernelWave
from ase.io.aff import affopen


def actually_calculate_kernel(*, gs, xcflags, q_empty, tag, ecut_max, fd,
                              timer, Eg, wd):
    kd = gs.kd
    bzq_qc = kd.get_bz_q_points(first=True)
    U_scc = kd.symmetry.op_scc
    ibzq_qc = kd.get_ibz_q_points(bzq_qc, U_scc)[0]

    l_l = np.array([1.0])

    if xcflags.linear_kernel:
        l_l = None
        omega_w = None
    elif not xcflags.dyn_kernel:
        omega_w = None
    else:
        omega_w = wd.omega_w

    kernel = KernelWave(
        l_l=l_l,
        omega_w=omega_w,
        gs=gs,
        xc=xcflags.xc,
        ibzq_qc=ibzq_qc,
        fd=fd,
        q_empty=q_empty,
        Eg=Eg,
        ecut=ecut_max,
        tag=tag,
        timer=timer)

    kernel.calculate_fhxc()


def calculate_kernel(*, ecut, xcflags, gs, nG, ns, iq, cut_G=None,
                     timer, fd, wd, Eg):
    xc = xcflags.xc
    tag = gs.atoms.get_chemical_formula(mode='hill')

    ecut_max = ecut * Ha  # XXX very ugly this
    q_empty = None

    filename = 'fhxc_%s_%s_%s_%s.ulm' % (tag, xc, ecut_max, iq)

    if not os.path.isfile(filename):
        q_empty = iq

    if xc != 'RPA':
        if q_empty is not None:
            actually_calculate_kernel(q_empty=q_empty, tag=tag,
                                      xcflags=xcflags, Eg=Eg,
                                      ecut_max=ecut_max, gs=gs,
                                      fd=fd, timer=timer,
                                      wd=wd)
            # (This creates the ulm file above.  Probably.)

        mpi.world.barrier()

        if xcflags.spin_kernel:
            with affopen(filename) as r:
                fv = r.fhxc_sGsG

            if cut_G is not None:
                cut_sG = np.tile(cut_G, ns)
                cut_sG[len(cut_G):] += len(fv) // ns
                fv = fv.take(cut_sG, 0).take(cut_sG, 1)

        else:
            if xc == 'RPA':
                fv = np.eye(nG)
            elif xc == 'range_RPA':
                raise NotImplementedError
#                    fv = np.exp(-0.25 * (G_G * self.range_rc) ** 2.0)

            elif xcflags.linear_kernel:
                with affopen(filename) as r:
                    fv = r.fhxc_sGsG

                if cut_G is not None:
                    fv = fv.take(cut_G, 0).take(cut_G, 1)

            elif not xcflags.dyn_kernel:
                # static kernel which does not scale with lambda
                with affopen(filename) as r:
                    fv = r.fhxc_lGG

                if cut_G is not None:
                    fv = fv.take(cut_G, 1).take(cut_G, 2)

            else:  # dynamical kernel
                with affopen(filename) as r:
                    fv = r.fhxc_lwGG

                if cut_G is not None:
                    fv = fv.take(cut_G, 2).take(cut_G, 3)
    else:
        fv = np.eye(nG)

    return fv



import matplotlib.pyplot as plt
from ase.units import Bohr
from gpaw.utilities.ps2ae import PS2AE
from gpaw import GPAW

calc = GPAW('hli.gpw', txt=None)

# Avarage PS potentials:
vh, vli = calc.get_atomic_electrostatic_potentials()
zh, zli = calc.atoms.positions[:, 2]
rh, rli = [max(setup.rcut_j) * Bohr for setup in calc.setups]
plt.plot([zh - rh, zh + rh], [vh, vh], label=r'$\tilde v$(H)')
plt.plot([zli - rli, zli + rli], [vli, vli], label=r'$\tilde v$(Li)')

# Transformer:
t = PS2AE(calc, grid_spacing=0.05)

# Interpolated PS and AE potentials:
ps = t.get_electrostatic_potential(ae=False)
ae = t.get_electrostatic_potential()
i = ps.shape[0] // 2
z = t.gd.coords(2) * Bohr

plt.plot(z, ps[i, i], '--', label=r'$\tilde v$')
plt.plot(z, ae[i, i], '-', label=r'$v$')

# Raw PS potential:
ps0 = calc.get_electrostatic_potential()
gd = calc.hamiltonian.finegd
i = ps0.shape[0] // 2
Z = gd.coords(2) * Bohr
plt.plot(Z, ps0[i, i], 'o')

plt.plot(z, 0 * z, 'k')
plt.xlabel('z [Ang]')
plt.ylabel('potential [eV]')
plt.ylim(bottom=-100, top=10)
plt.legend()
plt.savefig('hli-pot.png')



from myqueue.workflow import run


def workflow():
    runs = [run(script='PES_CO.py', cores=8, tmax='1h'),
            run(script='PES_H2O.py', cores=8, tmax='1h'),
            run(script='PES_NH3.py', cores=8, tmax='55m')]
    run(script='PES_plot.py', deps=runs)



import numpy as np
from gpaw.utilities.extrapolate import extrapolate

CO_rpa = np.loadtxt('CO.ralda_rpa_CO.dat')
C_rpa = np.loadtxt('CO.ralda_rpa_C.dat')
O_rpa = np.loadtxt('CO.ralda_rpa_O.dat')

a = CO_rpa
a[:, 1] -= (C_rpa[:, 1] + O_rpa[:, 1])
ext, A, B, sigma = extrapolate(a[:, 0], a[:, 1], reg=3, plot=False)

CO_rapbe = np.loadtxt('CO.ralda_rapbe_CO.dat')
C_rapbe = np.loadtxt('CO.ralda_rapbe_C.dat')
O_rapbe = np.loadtxt('CO.ralda_rapbe_O.dat')

a = CO_rapbe
a[:, 1] -= (C_rapbe[:, 1] + O_rapbe[:, 1])
ext, A, B, sigma = extrapolate(a[:, 0], a[:, 1], reg=3, plot=False)



from ase import Atom, io, optimize
from gpaw import GPAW, FermiDirac
from gpaw.cluster import Cluster
from gpaw.lrtddft import LrTDDFT
from gpaw.lrtddft.excited_state import ExcitedState

box = 5.      # box dimension
h = 0.25      # grid spacing
width = 0.01  # Fermi width
nbands = 6    # bands in GS calculation
nconv = 4     # bands in GS calculation to converge
R = 2.99      # starting distance
iex = 1       # excited state index
d = 0.01      # step for numerical force evaluation
exc = 'LDA'   # xc for the linear response TDDFT kernel
parallel = 4  # number of independent parrallel workers for force evaluation

s = Cluster([Atom('Na'), Atom('Na', [0, 0, R])])
s.minimal_box(box, h=h)

c = GPAW(h=h, nbands=nbands, eigensolver='cg',
         occupations=FermiDirac(width=width),
         setups={'Na': '1'},
         convergence={'bands': nconv})
c.calculate(s)
lr = LrTDDFT(c, xc=exc, restrict={'eps': 0.1, 'jend': nconv - 1})

ex = ExcitedState(lr, iex, d=d, parallel=parallel)
s.calc = ex

ftraj = 'relax_ex' + str(iex)
ftraj += '_box' + str(box) + '_h' + str(h)
ftraj += '_d' + str(d) + '.traj'
traj = io.Trajectory(ftraj, 'w', s)
dyn = optimize.FIRE(s)
dyn.attach(traj.write)
dyn.run(fmax=0.05)



import numpy as np
from gpaw.defects import ElectrostaticCorrections

sigma = 2 / (2.0 * np.sqrt(2.0 * np.log(2.0)))
q = -3
epsilon = 12.7
corrected = []
uncorrected = []
repeats = [1, 2, 3, 4]
for repeat in repeats:
    pristine = 'GaAs{0}{0}{0}.pristine.gpw'.format(repeat)
    charged = 'GaAs{0}{0}{0}.Ga_vac_charged.gpw'.format(repeat)
    elc = ElectrostaticCorrections(pristine=pristine,
                                   charged=charged,
                                   q=q,
                                   sigma=sigma,
                                   dimensionality='3d')
    elc.set_epsilons(epsilon)
    corrected.append(elc.calculate_corrected_formation_energy())
    uncorrected.append(elc.calculate_uncorrected_formation_energy())
    data = elc.collect_electrostatic_data()
    np.savez('electrostatic_data_{0}{0}{0}.npz'.format(repeat), **data)

np.savez('formation_energies.npz',
         repeats=np.array(repeats),
         corrected=np.array(corrected),
         uncorrected=np.array(uncorrected))



from ase import Atoms
from gpaw import GPAW

h = 0.2
n = 24
a = n * h
b = a / 2
H = Atoms('H', [(b - 0.1, b, b)], pbc=True, cell=(a, a, a))
calc = GPAW(nbands=1, gpts=(n, n, n), symmetry='off', txt='ltt.txt')
H.calc = calc
e0 = H.get_potential_energy()
for i in range(50):
    e = H.get_potential_energy()
    H.positions += (0.09123456789, 0.0423456789, 0.03456789)
assert abs(e - e0) < 0.0006
print(e, e0, e - e0)



from gpaw import GPAW, PW, MethfesselPaxton
from ase.spacegroup import crystal
from ase.io import write

a = 4.55643
mnsi = crystal(['Mn', 'Si'],
               [(0.1380, 0.1380, 0.1380), (0.84620, 0.84620, 0.84620)],
               spacegroup=198,
               cellpar=[a, a, a, 90, 90, 90])


for atom in mnsi:
    if atom.symbol == 'Mn':
        atom.magmom = 0.5

mnsi.calc = GPAW(xc='PBE',
                 kpts=(2, 2, 2),
                 mode=PW(800),
                 occupations=MethfesselPaxton(width=0.005),
                 txt='mnsi.txt')

mnsi.get_potential_energy()
mnsi.calc.write('mnsi.gpw')
v = mnsi.calc.get_electrostatic_potential()
write('mnsi.cube', mnsi, data=v)

assert abs(v.max() - 13.43) < 0.01



# creates: metric.png

import numpy as np
import matplotlib.pyplot as plt
from math import pi, cos


# Special points in the BZ of a simple cubic cell
G = pi * np.array([0., 0., 0.])
R = pi * np.array([1., 1., 1.])
X = pi * np.array([1., 0., 0.])
M = pi * np.array([1., 1., 0.])

# The path for the band plot
path = [X, G, R, X, M, G]
textpath = [r'$X$', r'$\Gamma$', r'$R$', r'$X$', r'$M$', r'$\Gamma$']

# Make band data
qvec = []
lines = [0]
previous = path[0]
for next in path[1:]:
    Npoints = int(round(20 * np.linalg.norm(next - previous)))
    lines.append(lines[-1] + Npoints)
    for t in np.linspace(0, 1, Npoints):
        qvec.append((1 - t) * previous + t * next)
    previous = next

vasp = [1 / max(np.linalg.norm(q), 1e-6)**2 for q in qvec]
gpaw = [(1 + cos(qx) + cos(qy) + cos(qz) +
         cos(qx) * cos(qy) + cos(qx) * cos(qz) + cos(qy) * cos(qz) +
         cos(qx) * cos(qy) * cos(qz)) / 8. for qx, qy, qz in qvec]

# Plot band data
fig = plt.figure(1, figsize=(5, 3), dpi=90)
fig.subplots_adjust(left=.1, right=.95)
lim = [0, lines[-1], 0, 1.25]
plt.plot(vasp, 'k:', label='VASP')
plt.plot(gpaw, 'k-', label='GPAW')
for q in lines:
    plt.plot([q, q], lim[2:], 'k-')
plt.xticks(lines, textpath)
plt.yticks([0, 1], [r'$1$', r'$w+1$'])
plt.axis(lim)

kwpad = {'borderpad': 0.2, 'borderaxespad': 0.06}

plt.legend(loc='upper right', **kwpad)
plt.title('Special metric for density changes')
plt.savefig('metric.png', dpi=90)
# plt.show()



import pytest
from ase import Atoms

from gpaw import GPAW, PW
from gpaw.mpi import world
from gpaw.xc.fxc import FXCCorrelation


@pytest.mark.response
def test_ralda_ralda_energy_H2(in_tmp_dir, scalapack):
    if world.size == 1:
        scalapack1 = None
        scalapack2 = None
    elif world.size == 2:
        scalapack1 = (2, world.size // 2, 32)
        scalapack2 = None
    else:
        scalapack1 = (2, world.size // 2, 32)
        scalapack2 = (2, world.size // 2, 32)

    # H2
    H2 = Atoms('H2', [(0, 0, 0), (0, 0, 0.7413)])
    H2.set_pbc(True)
    H2.set_cell((2., 2., 3.))
    H2.center()
    calc = GPAW(mode=PW(210, force_complex_dtype=True),
                eigensolver='rmm-diis',
                xc='LDA',
                basis='dzp',
                nbands=8,
                parallel={'domain': 1},
                convergence={'density': 1.e-6})
    H2.calc = calc
    H2.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=80, scalapack=scalapack1)
    calc.write('H2.gpw', mode='all')

    ralda = FXCCorrelation('H2.gpw', xc='rALDA', nblocks=min(4, world.size))
    E_ralda_H2 = ralda.calculate(ecut=[200])

    rapbe = FXCCorrelation('H2.gpw', xc='rAPBE')
    E_rapbe_H2 = rapbe.calculate(ecut=[200])

    # H
    H = Atoms('H')
    H.set_pbc(True)
    H.set_cell((2., 2., 3.))
    H.center()
    calc = GPAW(mode=PW(210, force_complex_dtype=True),
                eigensolver='rmm-diis',
                xc='LDA',
                basis='dzp',
                nbands=4,
                hund=True,
                parallel={'domain': 1},
                convergence={'density': 1.e-6})
    H.calc = calc
    H.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=80, scalapack=scalapack2)
    calc.write('H.gpw', mode='all')

    ralda = FXCCorrelation('H.gpw', xc='rALDA')
    E_ralda_H = ralda.calculate(ecut=[200])

    rapbe = FXCCorrelation('H.gpw', xc='rAPBE', nblocks=min(4, world.size))
    E_rapbe_H = rapbe.calculate(ecut=[200])

    assert E_ralda_H2 == pytest.approx(-0.8411, abs=0.001)
    assert E_ralda_H == pytest.approx(0.0029, abs=0.0001)
    assert E_rapbe_H2 == pytest.approx(-0.7233, abs=0.001)
    assert E_rapbe_H == pytest.approx(0.0161, abs=0.0001)



"""DFT ground state calculation of Fe(bcc), storing occupied and unoccupied
Kohn-Sham orbitals in a file."""

# Script modules
from gpaw import GPAW, PW, FermiDirac
from ase.build import bulk

# ---------- Inputs ---------- #

# Crystal structure
a = 2.867  # Lattice constant
mm = 2.21  # Initial magnetic moment

# The derivation of the Heisenberg exchange constants based on the magnetic
# force theorem takes an outset in the L(S)DA exchange-correlation functional.
# Thus, to be formally consistent, we need to use the LDA functional for the
# ground state.
xc = 'LDA'
# The plane wave energy cutoff of the ground state calculation is chosen:
# 1) high enough to provide well converged Kohn-Sham orbitals,
# 2) larger or equal to the energy cutoff of the mft response calculation.
# In a benchmark study, the latter will typically be the stricter requirement.
pw = 800  # eV
# When specifying the k-point grid, one should keep in mind that the q-points
# of the mft response calculation are required to be commensurate with the
# grid. For a bcc crystal, a k-point grid that samples a multiple of 4 k-points
# along all axes will support calculations at the high-symmetry points N, H and
# P. We choose N_k^(1/3)=32, which should yield converged magnon energies
# according to the convergence study of [arXiv:2204.04169].
kpts = 32  # final grid is (kpts, kpts, kpts)
# In the original paper [arXiv:2204.04169], it was shown that the mft response
# calculations are well converged when including only bands corresponding to
# shells with partial or full occupation.
nbands_mft = 6  # 4s + 3d
# To stabilize the ground state calculation, we add some extra empty bands
nbands_gs = nbands_mft + 4  # 4p + 5s
# For response calculations, we typically need a more strict convergence of the
# ground state density, magnetization and Kohn-Sham orbitals than usual.
conv = {'density': 1.e-8,
        'forces': 1.e-8,
        'bands': nbands_mft}  # Converge only the bands needed subsequently
occw = 0.001  # Fermi temperature in eV

# ---------- Script ---------- #

# Set up crystal structure
atoms = bulk('Fe', 'bcc', a=a)
atoms.set_initial_magnetic_moments([mm])
atoms.center()

# Construct the ground state calculator
kpt_grid = {'size': (kpts, kpts, kpts),
            'gamma': True}  # When converged, the grid offset shouldn't matter
calc = GPAW(xc='LDA',
            mode=PW(pw),
            kpts=kpt_grid,
            nbands=nbands_gs,
            convergence=conv,
            occupations=FermiDirac(occw),
            txt='Fe_gs.txt')

# DFT ground state calculation
atoms.calc = calc
atoms.get_potential_energy()

# Save the ground state and Kohn-Sham orbitals
# WARNING: This will generally leave large files on your computer. It is wise
# to remove them, once all subsequent calculations have finished.
calc.write('Fe_all.gpw', mode='all')



import numbers

import numpy as np

from ase.utils.timing import timer

import gpaw.mpi as mpi
from gpaw.fd_operators import Gradient
from gpaw.response.pw_parallelization import block_partition
from gpaw.response.symmetry import KPointFinder
from gpaw.response.context import calc_and_context
from gpaw.utilities.blas import mmm


class KPoint:
    def __init__(self, s, K, n1, n2, blocksize, na, nb,
                 ut_nR, eps_n, f_n, P_ani, shift_c):
        self.s = s    # spin index
        self.K = K    # BZ k-point index
        self.n1 = n1  # first band
        self.n2 = n2  # first band not included
        self.blocksize = blocksize
        self.na = na  # first band of block
        self.nb = nb  # first band of block not included
        self.ut_nR = ut_nR      # periodic part of wave functions in real-space
        self.eps_n = eps_n      # eigenvalues
        self.f_n = f_n          # occupation numbers
        self.P_ani = P_ani      # PAW projections
        self.shift_c = shift_c  # long story - see the
        # PairDensity.construct_symmetry_operators() method


class PairDistribution:
    def __init__(self, pair, mysKn1n2):
        self.pair = pair
        self.mysKn1n2 = mysKn1n2
        self.mykpts = [self.pair.get_k_point(s, K, n1, n2)
                       for s, K, n1, n2 in self.mysKn1n2]

    def kpt_pairs_by_q(self, q_c, m1, m2):
        pair = self.pair
        mykpts = self.mykpts
        for u, kpt1 in enumerate(mykpts):
            progress = u / len(mykpts)
            K2 = pair.kd.find_k_plus_q(q_c, [kpt1.K])[0]
            kpt2 = pair.get_k_point(kpt1.s, K2, m1, m2, block=True)

            yield progress, kpt1, kpt2


class KPointPair:
    """This class defines the kpoint-pair container object.

    Used for calculating pair quantities it contains two kpoints,
    and an associated set of Fourier components."""
    def __init__(self, kpt1, kpt2, Q_G):
        self.kpt1 = kpt1
        self.kpt2 = kpt2
        self.Q_G = Q_G

    def get_k1(self):
        """ Return KPoint object 1."""
        return self.kpt1

    def get_k2(self):
        """ Return KPoint object 2."""
        return self.kpt2

    def get_planewave_indices(self):
        """ Return the planewave indices associated with this pair."""
        return self.Q_G

    def get_transition_energies(self, n_n, m_m):
        """Return the energy difference for specified bands."""
        n_n = np.array(n_n)
        m_m = np.array(m_m)
        kpt1 = self.kpt1
        kpt2 = self.kpt2
        deps_nm = (kpt1.eps_n[n_n - self.kpt1.n1][:, np.newaxis] -
                   kpt2.eps_n[m_m - self.kpt2.n1])
        return deps_nm

    def get_occupation_differences(self, n_n, m_m):
        """Get difference in occupation factor between specified bands."""
        n_n = np.array(n_n)
        m_m = np.array(m_m)
        kpt1 = self.kpt1
        kpt2 = self.kpt2
        df_nm = (kpt1.f_n[n_n - self.kpt1.n1][:, np.newaxis] -
                 kpt2.f_n[m_m - self.kpt2.n1])
        return df_nm


class NoCalculatorPairDensity:
    def __init__(self, gs, *, context, ftol=1e-6,
                 threshold=1, real_space_derivatives=False, nblocks=1):
        self.gs = gs
        self.context = context

        self.fd = context.fd
        self.timer = context.timer
        self.world = context.world

        assert self.gs.kd.symmetry.symmorphic

        self.ftol = ftol
        self.threshold = threshold
        self.real_space_derivatives = real_space_derivatives

        self.blockcomm, self.kncomm = block_partition(context.world, nblocks)
        self.nblocks = nblocks

        self.fermi_level = self.gs.fermi_level
        self.spos_ac = self.gs.spos_ac

        self.nocc1 = None  # number of completely filled bands
        self.nocc2 = None  # number of non-empty bands
        self.count_occupied_bands()

        self.ut_sKnvR = None  # gradient of wave functions for optical limit

        self.vol = self.gs.gd.volume

        self.kd = self.gs.kd
        self.kptfinder = KPointFinder(self.kd.bzk_kc)
        print('Number of blocks:', nblocks, file=self.fd)

    def find_kpoint(self, k_c):
        return self.kptfinder.find(k_c)

    def count_occupied_bands(self):
        self.nocc1 = 9999999
        self.nocc2 = 0
        for kpt in self.gs.kpt_u:
            f_n = kpt.f_n / kpt.weight
            self.nocc1 = min((f_n > 1 - self.ftol).sum(), self.nocc1)
            self.nocc2 = max((f_n > self.ftol).sum(), self.nocc2)
        print('Number of completely filled bands:', self.nocc1, file=self.fd)
        print('Number of non-empty bands:', self.nocc2, file=self.fd)
        print('Total number of bands:', self.gs.bd.nbands,
              file=self.fd)

    def distribute_k_points_and_bands(self, band1, band2, kpts=None):
        """Distribute spins, k-points and bands.

        The attribute self.mysKn1n2 will be set to a list of (s, K, n1, n2)
        tuples that this process handles.
        """

        gs = self.gs

        if kpts is None:
            kpts = np.arange(gs.kd.nbzkpts)

        # nbands is the number of bands for each spin/k-point combination.
        nbands = band2 - band1
        size = self.kncomm.size
        rank = self.kncomm.rank
        ns = gs.nspins
        nk = len(kpts)
        n = (ns * nk * nbands + size - 1) // size
        i1 = min(rank * n, ns * nk * nbands)
        i2 = min(i1 + n, ns * nk * nbands)

        mysKn1n2 = []
        i = 0
        for s in range(ns):
            for K in kpts:
                n1 = min(max(0, i1 - i), nbands)
                n2 = min(max(0, i2 - i), nbands)
                if n1 != n2:
                    mysKn1n2.append((s, K, n1 + band1, n2 + band1))
                i += nbands

        print('BZ k-points:', gs.kd, file=self.fd)
        print('Distributing spins, k-points and bands (%d x %d x %d)' %
              (ns, nk, nbands),
              'over %d process%s' %
              (self.kncomm.size, ['es', ''][self.kncomm.size == 1]),
              file=self.fd)
        print('Number of blocks:', self.blockcomm.size, file=self.fd)

        return PairDistribution(self, mysKn1n2)

    @timer('Get a k-point')
    def get_k_point(self, s, k_c, n1, n2, load_wfs=True, block=False):
        """Return wave functions for a specific k-point and spin.

        s: int
            Spin index (0 or 1).
        K: int
            BZ k-point index.
        n1, n2: int
            Range of bands to include.
        """

        assert n1 <= n2

        gs = self.gs
        kd = gs.kd

        # Parse kpoint: is k_c an index or a vector
        if not isinstance(k_c, numbers.Integral):
            K = self.kptfinder.find(k_c)
            shift0_c = (kd.bzk_kc[K] - k_c).round().astype(int)
        else:
            # Fall back to index
            K = k_c
            shift0_c = np.array([0, 0, 0])
            k_c = None

        if block:
            nblocks = self.blockcomm.size
            rank = self.blockcomm.rank
        else:
            nblocks = 1
            rank = 0

        blocksize = (n2 - n1 + nblocks - 1) // nblocks
        na = min(n1 + rank * blocksize, n2)
        nb = min(na + blocksize, n2)

        U_cc, T, a_a, U_aii, shift_c, time_reversal = \
            self.construct_symmetry_operators(K, k_c=k_c)

        shift_c += -shift0_c
        ik = kd.bz2ibz_k[K]
        assert kd.comm.size == 1
        kpt = gs.kpt_qs[ik][s]

        assert n2 <= len(kpt.eps_n), \
            'Increase GS-nbands or decrease chi0-nbands!'
        eps_n = kpt.eps_n[n1:n2]
        f_n = kpt.f_n[n1:n2] / kpt.weight

        if not load_wfs:
            return KPoint(s, K, n1, n2, blocksize, na, nb,
                          None, eps_n, f_n, None, shift_c)

        with self.timer('load wfs'):
            psit_nG = kpt.psit_nG
            ut_nR = gs.gd.empty(nb - na, gs.dtype)
            for n in range(na, nb):
                ut_nR[n - na] = T(gs.pd.ifft(psit_nG[n], ik))

        with self.timer('Load projections'):
            P_ani = []
            for b, U_ii in zip(a_a, U_aii):
                P_ni = np.dot(kpt.P_ani[b][na:nb], U_ii)
                if time_reversal:
                    P_ni = P_ni.conj()
                P_ani.append(P_ni)

        return KPoint(s, K, n1, n2, blocksize, na, nb,
                      ut_nR, eps_n, f_n, P_ani, shift_c)

    @timer('Get kpoint pair')
    def get_kpoint_pair(self, pd, s, Kork_c, n1, n2, m1, m2,
                        load_wfs=True, block=False):
        assert m1 <= m2
        assert n1 <= n2

        if isinstance(Kork_c, int):
            # If k_c is an integer then it refers to
            # the index of the kpoint in the BZ
            k_c = self.gs.kd.bzk_kc[Kork_c]
        else:
            k_c = Kork_c

        q_c = pd.kd.bzk_kc[0]
        with self.timer('get k-points'):
            kpt1 = self.get_k_point(s, k_c, n1, n2, load_wfs=load_wfs)
            # K2 = wfs.kd.find_k_plus_q(q_c, [kpt1.K])[0]
            kpt2 = self.get_k_point(s, k_c + q_c, m1, m2,
                                    load_wfs=load_wfs, block=block)

        with self.timer('fft indices'):
            Q_G = self.get_fft_indices(kpt1.K, kpt2.K, q_c, pd,
                                       kpt1.shift_c - kpt2.shift_c)

        return KPointPair(kpt1, kpt2, Q_G)

    @timer('get_pair_density')
    def get_pair_density(self, pd, kptpair, n_n, m_m,
                         optical_limit=False, intraband=False,
                         Q_aGii=None, block=False, direction=2,
                         extend_head=True):
        """Get pair density for a kpoint pair."""
        ol = optical_limit = np.allclose(pd.kd.bzk_kc[0], 0.0)
        eh = extend_head
        cpd = self.calculate_pair_densities  # General pair densities
        opd = self.optical_pair_density  # Interband pair densities / q

        if Q_aGii is None:
            Q_aGii = self.initialize_paw_corrections(pd)

        kpt1 = kptpair.kpt1
        kpt2 = kptpair.kpt2
        Q_G = kptpair.Q_G  # Fourier components of kpoint pair
        nG = len(Q_G)

        if extend_head:
            n_nmG = np.zeros((len(n_n), len(m_m), nG + 2 * ol), pd.dtype)
        else:
            n_nmG = np.zeros((len(n_n), len(m_m), nG), pd.dtype)

        for j, n in enumerate(n_n):
            Q_G = kptpair.Q_G
            with self.timer('conj'):
                ut1cc_R = kpt1.ut_nR[n - kpt1.na].conj()
            with self.timer('paw'):
                C1_aGi = [np.dot(Q_Gii, P1_ni[n - kpt1.na].conj())
                          for Q_Gii, P1_ni in zip(Q_aGii, kpt1.P_ani)]
                n_nmG[j, :, 2 * ol * eh:] = cpd(ut1cc_R, C1_aGi, kpt2, pd, Q_G,
                                                block=block)
            if optical_limit:
                if extend_head:
                    n_nmG[j, :, 0:3] = opd(n, m_m, kpt1, kpt2,
                                           block=block)
                else:
                    n_nmG[j, :, 0] = opd(n, m_m, kpt1, kpt2,
                                         block=block)[:, direction]
        return n_nmG

    @timer('get_pair_momentum')
    def get_pair_momentum(self, pd, kptpair, n_n, m_m, Q_avGii=None):
        r"""Calculate matrix elements of the momentum operator.

        Calculates::

          n_{nm\mathrm{k}}\int_{\Omega_{\mathrm{cell}}}\mathrm{d}\mathbf{r}
          \psi_{n\mathrm{k}}^*(\mathbf{r})
          e^{-i\,(\mathrm{q} + \mathrm{G})\cdot\mathbf{r}}
          \nabla\psi_{m\mathrm{k} + \mathrm{q}}(\mathbf{r})

        pd: PlaneWaveDescriptor
            Plane wave descriptor of a single q_c.
        kptpair: KPointPair
            KpointPair object containing the two kpoints.
        n_n: list
            List of left-band indices (n).
        m_m:
            List of right-band indices (m).
        """
        gs = self.gs

        kpt1 = kptpair.kpt1
        kpt2 = kptpair.kpt2
        Q_G = kptpair.Q_G  # Fourier components of kpoint pair

        # For the same band we
        kd = gs.kd
        gd = gs.gd
        k_c = kd.bzk_kc[kpt1.K] + kpt1.shift_c
        k_v = 2 * np.pi * np.dot(k_c, np.linalg.inv(gd.cell_cv).T)

        # Calculate k + G
        G_Gv = pd.get_reciprocal_vectors(add_q=True)
        kqG_Gv = k_v[np.newaxis] + G_Gv

        # Pair velocities
        n_nmvG = pd.zeros((len(n_n), len(m_m), 3))

        # Calculate derivatives of left-wavefunction
        # (there will typically be fewer of these)
        ut_nvR = self.make_derivative(kpt1.s, kpt1.K, kpt1.n1, kpt1.n2)

        # PAW-corrections
        if Q_avGii is None:
            Q_avGii = self.initialize_paw_nabla_corrections(pd)

        # Iterate over occupied bands
        for j, n in enumerate(n_n):
            ut1cc_R = kpt1.ut_nR[n].conj()

            n_mG = self.calculate_pair_densities(ut1cc_R,
                                                 [], kpt2,
                                                 pd, Q_G)

            n_nmvG[j] = 1j * kqG_Gv.T[np.newaxis] * n_mG[:, np.newaxis]

            # Treat each cartesian component at a time
            for v in range(3):
                # Minus from integration by parts
                utvcc_R = -ut_nvR[n, v].conj()
                Cv1_aGi = [np.dot(P1_ni[n].conj(), Q_vGii[v])
                           for Q_vGii, P1_ni in zip(Q_avGii, kpt1.P_ani)]

                nv_mG = self.calculate_pair_densities(utvcc_R,
                                                      Cv1_aGi, kpt2,
                                                      pd, Q_G)

                n_nmvG[j, :, v] += nv_mG

        # We want the momentum operator
        n_nmvG *= -1j

        return n_nmvG

    @timer('Calculate pair-densities')
    def calculate_pair_densities(self, ut1cc_R, C1_aGi, kpt2, pd, Q_G,
                                 block=True):
        """Calculate FFT of pair-densities and add PAW corrections.

        ut1cc_R: 3-d complex ndarray
            Complex conjugate of the periodic part of the left hand side
            wave function.
        C1_aGi: list of ndarrays
            PAW corrections for all atoms.
        kpt2: KPoint object
            Right hand side k-point object.
        pd: PWDescriptor
            Plane-wave descriptor for for q=k2-k1.
        Q_G: 1-d int ndarray
            Mapping from flattened 3-d FFT grid to 0.5(G+q)^2<ecut sphere.
        """

        dv = pd.gd.dv
        n_mG = pd.empty(kpt2.blocksize)
        myblocksize = kpt2.nb - kpt2.na

        for ut_R, n_G in zip(kpt2.ut_nR, n_mG):
            n_R = ut1cc_R * ut_R
            with self.timer('fft'):
                n_G[:] = pd.fft(n_R, 0, Q_G) * dv
        # PAW corrections:
        with self.timer('gemm'):
            for C1_Gi, P2_mi in zip(C1_aGi, kpt2.P_ani):
                # gemm(1.0, C1_Gi, P2_mi, 1.0, n_mG[:myblocksize], 't')
                mmm(1.0, P2_mi, 'N', C1_Gi, 'T', 1.0, n_mG[:myblocksize])

        if not block or self.blockcomm.size == 1:
            return n_mG
        else:
            n_MG = pd.empty(kpt2.blocksize * self.blockcomm.size)
            self.blockcomm.all_gather(n_mG, n_MG)
            return n_MG[:kpt2.n2 - kpt2.n1]

    @timer('Optical limit')
    def optical_pair_velocity(self, n, m_m, kpt1, kpt2, block=False):
        if self.ut_sKnvR is None or kpt1.K not in self.ut_sKnvR[kpt1.s]:
            self.ut_sKnvR = self.calculate_derivatives(kpt1)

        kd = self.gs.kd
        gd = self.gs.gd
        k_c = kd.bzk_kc[kpt1.K] + kpt1.shift_c
        k_v = 2 * np.pi * np.dot(k_c, np.linalg.inv(gd.cell_cv).T)

        ut_vR = self.ut_sKnvR[kpt1.s][kpt1.K][n - kpt1.n1]
        atomdata_a = self.gs.setups
        C_avi = [np.dot(atomdata.nabla_iiv.T, P_ni[n - kpt1.na])
                 for atomdata, P_ni in zip(atomdata_a, kpt1.P_ani)]

        blockbands = kpt2.nb - kpt2.na
        n0_mv = np.empty((kpt2.blocksize, 3), dtype=complex)
        nt_m = np.empty(kpt2.blocksize, dtype=complex)
        n0_mv[:blockbands] = -self.gs.gd.integrate(ut_vR,
                                                   kpt2.ut_nR).T
        nt_m[:blockbands] = self.gs.gd.integrate(kpt1.ut_nR[n - kpt1.na],
                                                 kpt2.ut_nR)

        n0_mv[:blockbands] += (1j * nt_m[:blockbands, np.newaxis] *
                               k_v[np.newaxis, :])

        for C_vi, P_mi in zip(C_avi, kpt2.P_ani):
            # gemm(1.0, C_vi, P_mi, 1.0, n0_mv[:blockbands], 'c')
            mmm(1.0, P_mi, 'N', C_vi, 'C', 1.0, n0_mv[:blockbands])

        if block and self.blockcomm.size > 1:
            n0_Mv = np.empty((kpt2.blocksize * self.blockcomm.size, 3),
                             dtype=complex)
            self.blockcomm.all_gather(n0_mv, n0_Mv)
            n0_mv = n0_Mv[:kpt2.n2 - kpt2.n1]

        return -1j * n0_mv

    def optical_pair_density(self, n, m_m, kpt1, kpt2,
                             block=False):
        # Relative threshold for perturbation theory
        threshold = self.threshold

        eps1 = kpt1.eps_n[n - kpt1.n1]
        deps_m = (eps1 - kpt2.eps_n)[m_m - kpt2.n1]
        n0_mv = self.optical_pair_velocity(n, m_m, kpt1, kpt2,
                                           block=block)

        deps_m = deps_m.copy()
        deps_m[deps_m == 0.0] = np.inf

        smallness_mv = np.abs(-1e-3 * n0_mv / deps_m[:, np.newaxis])
        inds_mv = (np.logical_and(np.inf > smallness_mv,
                                  smallness_mv > threshold))
        n0_mv *= - 1 / deps_m[:, np.newaxis]
        n0_mv[inds_mv] = 0

        return n0_mv

    @timer('Intraband')
    def intraband_pair_density(self, kpt, n_n=None,
                               only_partially_occupied=False):
        """Calculate intraband matrix elements of nabla"""
        # Bands and check for block parallelization
        na, nb, n1 = kpt.na, kpt.nb, kpt.n1
        vel_nv = np.zeros((nb - na, 3), dtype=complex)
        if n_n is None:
            n_n = np.arange(na, nb)
        assert np.max(n_n) < nb, 'This is too many bands'
        assert np.min(n_n) >= na, 'This is too few bands'

        # Load kpoints
        kd = self.gs.kd
        gd = self.gs.gd
        k_c = kd.bzk_kc[kpt.K] + kpt.shift_c
        k_v = 2 * np.pi * np.dot(k_c, np.linalg.inv(gd.cell_cv).T)
        atomdata_a = self.gs.setups
        f_n = kpt.f_n

        width = self.gs.get_occupations_width()

        if width > 1e-15:
            dfde_n = -1 / width * (f_n - f_n**2.0)  # Analytical derivative
            partocc_n = np.abs(dfde_n) > 1e-5  # Is part. occupied?
        else:
            # Just include all bands to be sure
            partocc_n = np.ones(len(f_n), dtype=bool)

        if only_partially_occupied and not partocc_n.any():
            return None

        if only_partially_occupied:
            # Check for block par. consistency
            assert (partocc_n < nb).all(), \
                print('Include more unoccupied bands ', +
                      'or less block parr.', file=self.fd)

        # Break bands into degenerate chunks
        degchunks_cn = []  # indexing c as chunk number
        for n in n_n:
            inds_n = np.nonzero(np.abs(kpt.eps_n[n - n1] -
                                       kpt.eps_n) < 1e-5)[0] + n1

            # Has this chunk already been computed?
            oldchunk = any([n in chunk for chunk in degchunks_cn])
            if not oldchunk and \
               (partocc_n[n - n1] or not only_partially_occupied):
                assert all([ind in n_n for ind in inds_n]), \
                    print('\nYou are cutting over a degenerate band ' +
                          'using block parallelization.',
                          inds_n, n_n, file=self.fd)
                degchunks_cn.append((inds_n))

        # Calculate matrix elements by diagonalizing each block
        for ind_n in degchunks_cn:
            deg = len(ind_n)
            ut_nvR = self.gs.gd.zeros((deg, 3), complex)
            vel_nnv = np.zeros((deg, deg, 3), dtype=complex)
            # States are included starting from kpt.na
            ut_nR = kpt.ut_nR[ind_n - na]

            # Get derivatives
            for ind, ut_vR in zip(ind_n, ut_nvR):
                ut_vR[:] = self.make_derivative(kpt.s, kpt.K,
                                                ind, ind + 1)[0]

            # Treat the whole degenerate chunk
            for n in range(deg):
                ut_vR = ut_nvR[n]
                C_avi = [np.dot(atomdata.nabla_iiv.T, P_ni[ind_n[n] - na])
                         for atomdata, P_ni in zip(atomdata_a, kpt.P_ani)]

                nabla0_nv = -self.gs.gd.integrate(ut_vR, ut_nR).T
                nt_n = self.gs.gd.integrate(ut_nR[n], ut_nR)
                nabla0_nv += 1j * nt_n[:, np.newaxis] * k_v[np.newaxis, :]

                for C_vi, P_ni in zip(C_avi, kpt.P_ani):
                    # gemm(1.0, C_vi, P_ni[ind_n - na], 1.0, nabla0_nv, 'c')
                    mmm(1.0, P_ni[ind_n - na], 'N', C_vi, 'C', 1.0, nabla0_nv)

                vel_nnv[n] = -1j * nabla0_nv

            for iv in range(3):
                vel, _ = np.linalg.eig(vel_nnv[..., iv])
                vel_nv[ind_n - na, iv] = vel  # Use eigenvalues

        return vel_nv[n_n - na]

    def get_fft_indices(self, K1, K2, q_c, pd, shift0_c):
        """Get indices for G-vectors inside cutoff sphere."""
        kd = self.gs.kd
        N_G = pd.Q_qG[0]
        shift_c = (shift0_c +
                   (q_c - kd.bzk_kc[K2] + kd.bzk_kc[K1]).round().astype(int))
        if shift_c.any():
            n_cG = np.unravel_index(N_G, pd.gd.N_c)
            n_cG = [n_G + shift for n_G, shift in zip(n_cG, shift_c)]
            N_G = np.ravel_multi_index(n_cG, pd.gd.N_c, 'wrap')
        return N_G

    def construct_symmetry_operators(self, K, k_c=None):
        from gpaw.response.symmetry_ops import construct_symmetry_operators
        return construct_symmetry_operators(
            self.gs, K, k_c, apply_strange_shift=False)

    @timer('Initialize PAW corrections')
    def initialize_paw_corrections(self, pd, soft=False):
        from gpaw.response.paw import calculate_paw_corrections
        return calculate_paw_corrections(
            setups=self.gs.setups, pd=pd, soft=soft,
            spos_ac=self.spos_ac)

    @timer('Initialize PAW corrections')
    def initialize_paw_nabla_corrections(self, pd, soft=False):
        print('Initializing nabla PAW Corrections', file=self.fd)
        from gpaw.response.paw import calculate_paw_nabla_corrections
        return calculate_paw_nabla_corrections(
            setups=self.gs.setups, pd=pd, soft=soft,
            spos_ac=self.spos_ac)

    def calculate_derivatives(self, kpt):
        ut_sKnvR = [{}, {}]
        ut_nvR = self.make_derivative(kpt.s, kpt.K, kpt.n1, kpt.n2)
        ut_sKnvR[kpt.s][kpt.K] = ut_nvR

        return ut_sKnvR

    @timer('Derivatives')
    def make_derivative(self, s, K, n1, n2):
        gs = self.gs
        if self.real_space_derivatives:
            grad_v = [Gradient(gs.gd, v, 1.0, 4, complex).apply
                      for v in range(3)]

        U_cc, T, a_a, U_aii, shift_c, time_reversal = \
            self.construct_symmetry_operators(K)
        A_cv = gs.gd.cell_cv
        M_vv = np.dot(np.dot(A_cv.T, U_cc.T), np.linalg.inv(A_cv).T)
        ik = gs.kd.bz2ibz_k[K]
        assert gs.kd.comm.size == 1
        kpt = gs.kpt_qs[ik][s]
        psit_nG = kpt.psit_nG
        iG_Gv = 1j * gs.pd.get_reciprocal_vectors(q=ik, add_q=False)
        ut_nvR = gs.gd.zeros((n2 - n1, 3), complex)
        for n in range(n1, n2):
            for v in range(3):
                if self.real_space_derivatives:
                    ut_R = T(gs.pd.ifft(psit_nG[n], ik))
                    grad_v[v](ut_R, ut_nvR[n - n1, v],
                              np.ones((3, 2), complex))
                else:
                    ut_R = T(gs.pd.ifft(iG_Gv[:, v] * psit_nG[n], ik))
                    for v2 in range(3):
                        ut_nvR[n - n1, v2] += ut_R * M_vv[v, v2]

        return ut_nvR

    def __del__(self):
        self.context.close()


class PairDensity(NoCalculatorPairDensity):
    def __init__(self, gs, *,
                 world=mpi.world, txt='-', timer=None,
                 **kwargs):
        """Density matrix elements

        Parameters
        ----------
        ftol : float
            Threshold determining whether a band is completely filled
            (f > 1 - ftol) or completely empty (f < ftol).
        threshold : float
            Numerical threshold for the optical limit k dot p perturbation
            theory expansion.
        real_space_derivatives : bool
            Calculate nabla matrix elements (in the optical limit)
            using a real space finite difference approximation.
        """

        # note: gs is just called gs for historical reasons.
        # It's actually calc-or-filename union.

        self.calc, context = calc_and_context(gs, txt, world, timer)

        super().__init__(
            gs=self.calc.gs_adapter(),
            context=context,
            **kwargs)



from gpaw.atom.generator import Generator

# List of elements for which setups will be generated
elements = ['N']

# Fraction of Weizsacker
lambda_coeff = 1.0

# Fraction of Thomas-Fermi
gamma_coeff = 1.0

# Select optimum cutoff and grid
for symbol in elements:
    gpernode = 800
    if symbol == 'H':
        rcut = 0.9
    elif symbol in ['He' or 'Li']:
        rcut = 1.0
    elif symbol in ['Be', 'B', 'C', 'N', 'O', 'F', 'Ne']:
        rcut = 1.2
    elif symbol in ['Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar']:
        rcut = 1.4
    else:
        rcut = 1.0

    # If the lambda scaling is used change name to differentiate the setup
    name = f'lambda_{lambda_coeff}'

    # Use of Kinetic functional (minus the Tw contribution) inside the
    # xc definition
    pauliname = f'{gamma_coeff}_LDA_K_TF+1.0_LDA_X+1.0_LDA_C_PW'

    # Calculate OFDFT density
    g = Generator(symbol, xcname=pauliname, scalarrel=False,
                  orbital_free=True, tw_coeff=lambda_coeff,
                  gpernode=gpernode)

    g.run(exx=False,
          name=name,
          rcut=rcut,
          write_xml=True)



# web-page: convergence.db
import numpy as np
from ase import Atoms
from ase.build import hcp0001
from ase.constraints import FixAtoms
from ase.optimize import BFGSLineSearch
from gpaw import GPAW, PW, Davidson

a = 2.72
c = 1.58 * a
vacuum = 4.0


def adsorb(db, height=1.2, nlayers=3, nkpts=7, ecut=400):
    """Adsorb nitrogen in hcp-site on Ru(0001) surface.

    Do calculations for N/Ru(0001), Ru(0001) and a nitrogen atom
    if they have not already been done.

    db: Database
        Database for collecting results.
    height: float
        Height of N-atom above top Ru-layer.
    nlayers: int
        Number of Ru-layers.
    nkpts: int
        Use a (nkpts * nkpts) Monkhorst-Pack grid that includes the
        Gamma point.
    ecut: float
        Cutoff energy for plane waves.

    Returns height.
    """

    name = f'Ru{nlayers}-{nkpts}x{nkpts}-{ecut:.0f}'

    parameters = dict(mode=PW(ecut),
                      eigensolver=Davidson(niter=2),
                      poissonsolver={'dipolelayer': 'xy'},
                      kpts={'size': (nkpts, nkpts, 1), 'gamma': True},
                      xc='PBE')

    # N/Ru(0001):
    slab = hcp0001('Ru', a=a, c=c, size=(1, 1, nlayers))
    z = slab.positions[:, 2].max() + height
    x, y = np.dot([2 / 3, 2 / 3], slab.cell[:2, :2])
    slab.append('N')
    slab.positions[-1] = [x, y, z]
    slab.center(vacuum=vacuum, axis=2)  # 2: z-axis

    # Fix first nlayer atoms:
    slab.constraints = FixAtoms(indices=list(range(nlayers)))

    id = db.reserve(name=f'N/{nlayers}Ru(0001)', nkpts=nkpts, ecut=ecut)
    if id is not None:  # skip calculation if already done
        slab.calc = GPAW(txt='N' + name + '.txt',
                         **parameters)
        optimizer = BFGSLineSearch(slab, logfile='N' + name + '.opt')
        optimizer.run(fmax=0.01)
        height = slab.positions[-1, 2] - slab.positions[:-1, 2].max()
        del db[id]
        db.write(slab,
                 name=f'N/{nlayers}Ru(0001)', nkpts=nkpts, ecut=ecut,
                 height=height)

    # Clean surface (single point calculation):
    id = db.reserve(name=f'{nlayers}Ru(0001)', nkpts=nkpts, ecut=ecut)
    if id is not None:
        del slab[-1]  # remove nitrogen atom
        slab.calc = GPAW(txt=name + '.txt',
                         **parameters)
        slab.get_forces()
        del db[id]
        db.write(slab,
                 name=f'{nlayers}Ru(0001)', nkpts=nkpts, ecut=ecut)

    # Nitrogen atom:
    id = db.reserve(name='N-atom', ecut=ecut)
    if id is not None:
        # Create spin-polarized nitrogen atom:
        molecule = Atoms('N', magmoms=[3])
        molecule.center(vacuum=4.0)
        # Remove parameters that make no sense for an isolated atom:
        del parameters['kpts']
        del parameters['poissonsolver']
        # Calculate energy:
        molecule.calc = GPAW(txt=name + '.txt', **parameters)
        molecule.get_potential_energy()
        del db[id]
        db.write(molecule, name='N-atom', ecut=ecut)

    return height


def run():
    h = 1.2
    from ase.db import connect
    db = connect('convergence.db')
    for n in range(1, 10):
        h = adsorb(db, h, n, 7, 400)
    for k in range(4, 18):
        h = adsorb(db, h, 2, k, 400)
    for ecut in range(350, 801, 50):
        h = adsorb(db, h, 2, 7, ecut)


if __name__ == '__main__':
    run()



from ase.build import bulk
from gpaw import GPAW, Mixer, PW


def test_reuse_wfs():
    atoms0 = bulk('Si')
    atoms0.rattle(stdev=0.01, seed=17)  # Break symmetry

    results = []

    for mode in ['fd', 'pw']:
        for method in [None, 'paw', 'lcao']:
            atoms = atoms0.copy()

            kwargs = {}
            if mode == 'pw':
                if method == 'lcao':
                    continue  # Not implemented yet
                kwargs.update(mode=PW(400.0))

            calc = GPAW(mixer=Mixer(0.4, 5, 20.0),
                        basis='dzp' if method == 'lcao' else {},
                        experimental={'reuse_wfs_method': method},
                        xc='oldLDA',
                        kpts=[2, 2, 2],
                        **kwargs)
            atoms.calc = calc
            E1 = atoms.get_potential_energy()
            niter1 = calc.scf.niter
            atoms.rattle(stdev=0.05)
            E2 = atoms.get_potential_energy()
            niter2 = calc.scf.niter
            results.append([mode, method, E1, E2, niter1, niter2])

    if calc.wfs.world.rank == 0:
        for result in results:
            print(' '.join(str(x) for x in result))



# flake8: noqa
class Density:
    def __init__(self):
        self.gd = None
        self.mixer = None
        self.nt_sG = None
        self.nct = None

    def read(self, reader):
        self.gd = reader.get_grid_descriptor()
        self.mixer = reader.read('mixer')
        self.nt_sG = reader.get_array('...')

    def update(self, atoms, **kwargs):
        # Initialize stuff:
        if self.gd is None:
            self.gd = ...
        if self.mixer is None:
            self.mixer = Mixer(self.gd)
        if self.nct is None:
            self.nct = LFC(atoms)

        # Change stuff:
        if 'mixer' in kwargs:
            self.mixer = kwargs['mixer']

        # Update stuff:
        self.nct.set_positions(atoms)

    def allocate(self, lfc=True):
        if lfc:
            self.nct.allocate()

    def memory_estimate(self):
        ...

    def write(self, foo):
        ...



def factorial(x):
    """Return x!, where x is a non-negative integer"""
    if x < 2: return 1
    else: return x * factorial(x - 1)

def gcd(a, b):
    """Return greatest common divisor of a and b.
       Uses the euclidian algorithm
    """
    if b == 0: return a
    else: return gcd(b, a % b)

class Rational:
    """Class used to represent rational numbers as fractions, such that
       no precision is lost during calculation operations.
       Example usage with Numeric:
       import numpy as np
       from tools import Rational as Q

       n = np.zeros(4, 'O')                   array([0 , 0 , 0 , 0 ],'O')
       n[2:4] = [Q(35,12), Q(36,12)]           array([0 , 0 , 35./12 , 3 ],'O')
       24 * n                                  array([0 , 0 , 70 , 72 ],'O')
       np.multiply(n, Q(3,9))                 array([0 , 0 , 35./36 , 1 ],'O')
       
    """
    def __init__(self, nom=0, denom=1):
##        assert type(nom) == type(denom) == int

        # ensure that sign is in the nominator
        nom = cmp(denom, 0) * nom
        denom = abs(denom)

        # reduce fraction
        q = gcd(nom, denom)
        self.nom = nom / q
        self.denom = denom / q

    def __add__(self, x):
        if type(x) == float:
            return float(self) + x
        elif type(x) == int:
            x = Rational(x)
        nom = self.nom * x.denom + x.nom * self.denom
        denom = self.denom * x.denom
        return Rational(nom, denom)

    def __radd__(self, x):
        return self.__add__(x)

    def __mul__(self, x):
        if type(x) == float:
            return float(self) * x
        elif type(x) == int:
            x = Rational(x)
        return Rational(self.nom * x.nom, self.denom * x.denom)

    def __rmul__(self, x):
        return self.__mul__(x)

    def __neg__(self):
        return Rational(-self.nom, self.denom)

    def __pos__(self):
        return self.copy()

    def __sub__(self, x):
        return self.__add__(-x)

    def __rsub__(self, x):
        return -self.__sub__(x)
    
    def __div__(self, x):
        if type(x) == float:
            return float(self) / x
        elif type(x) == int:
            x = Rational(x)
        return self.__mul__(Rational(x.denom, x.nom))

    def __rdiv__(self, x):
        if type(x) == float:
            return x / float(self)
        elif type(x) == int:
            x = Rational(x)
        return x.__mul__(Rational(self.denom, self.nom))

    def __pow__(self, p):
        if p == 0: return Rational(1)
        if p >= 0 and type(p) == int:
            return Rational(self.nom**p, self.denom**p)
        else:
            return float(self)**p

    def __mod__(self, x):
        if type(x) == float:
            return float(self) % x
        return Rational(self.nom % (x * self.denom), self.denom)

    def __rmod__(self, x):
        if type(x) == int:
            x = Rational(x)
        i = self.__int__()
        return x.__mod__(i)

    def __abs__(self):
        return Rational(abs(self.nom), self.denom)
    
    def __nonzero__(self):
        return self.nom.__nonzero__()

    def __cmp__(self, x):
        return cmp(float(self), float(x))    
        
    def __str__(self):
        out = str(self.nom)
        if self.denom != 1:
            out += './' + str(self.denom)
        return out

    def __int__(self):
        assert self.denom == 1
        return self.nom

    def __float__(self):
        return float(self.nom) / self.denom

    def __repr__(self):
        out = repr(self.nom)
        if self.denom != 1:
            out += './' + repr(self.denom)
        return out

    def __copy__(self):
        return Rational(self.nom, self.denom)

    def floor(self):
        return int(float(self))

    def sqrt(self):
        return self**.5

    def abs(self):
        return Rational(abs(self.nom), self.denom)
    
    def copy(self):
        return Rational(self.nom, self.denom)



from ase import Atoms
from gpaw import GPAW, FermiDirac, Mixer
from gpaw import PW
from gpaw.hybrids.energy import non_self_consistent_energy as nsc_energy

from ase.parallel import paropen

# This calculation is too heavy to run as an exercise!!

myresults = paropen('si.atom.pbe_and_exx_energies.txt', 'a')

# Plane wave cutoff
pwcutoff = 400.0

# Do the isolated calculation

for L in [6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0]:
    isolated_silicon = Atoms(
        symbols=['Si'],
        positions=[[0.5 * L, 0.5 * L, 0.5 * L]],
        cell=([L, 0, 0], [0, L, 0], [0, 0, L]),
        pbc=(1, 1, 1))

    isolated_calc = GPAW(
        mode=PW(pwcutoff, force_complex_dtype=True),
        xc='PBE',
        txt='si_isolated_pbe.txt',
        occupations=FermiDirac(0.01, fixmagmom=True),
        spinpol=True,
        hund=True,
        convergence={'density': 1.e-6},
        mixer=Mixer(beta=0.05, nmaxold=5, weight=50.0))

    isolated_silicon.calc = isolated_calc

    e0_isolated_pbe = isolated_silicon.get_potential_energy()
    isolated_calc.write('si.pbe+exx.isolated.gpw', mode='all')

    # Now the exact exchange
    si_isolated_exx = nsc_energy('si.pbe+exx.isolated.gpw', 'EXX')

    s = str(L)
    s += ' '
    s += str(e0_isolated_pbe)
    s += ' '
    s += str(si_isolated_exx)
    s += '\n'
    myresults.write(s)



from ase.parallel import paropen
from ase.units import Hartree
from gpaw.xc.rpa import RPACorrelation

f = paropen('con_freq.dat', 'w')
for N in [4, 6, 8, 12, 16, 24, 32]:
    rpa = RPACorrelation('N2.gpw', txt='rpa_N2_frequencies.txt',
                         nfrequencies=N)
    E = rpa.calculate(ecut=[50])
    print(N, E[0], file=f)
    if N == 16:
        f16 = paropen('frequency_gauss16.dat', 'w')
        for w, e in zip(rpa.omega_w, rpa.E_w):
            print(w * Hartree, e, file=f16)
        f16.close()
f.close()



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

import numpy as np

import gpaw.fftw as fftw
from gpaw.core.domain import Domain
from gpaw.core.matrix import Matrix
from gpaw.mpi import MPIComm
from gpaw.typing import Array1D

if TYPE_CHECKING:
    from gpaw.core.uniform_grid import UniformGridFunctions, UniformGrid

from gpaw.new import prod

DomainType = TypeVar('DomainType', bound=Domain)


class DistributedArrays(Generic[DomainType]):
    desc: DomainType

    def __init__(self,
                 dims: int | tuple[int, ...],
                 myshape: tuple[int, ...],
                 comm: MPIComm,
                 domain_comm: MPIComm,
                 data: np.ndarray | None,
                 dv: float,
                 dtype):
        self.myshape = myshape
        self.comm = comm
        self.domain_comm = domain_comm
        self.dv = dv

        # convert int to tuple:
        self.dims = dims if isinstance(dims, tuple) else (dims,)

        if self.dims:
            mydims0 = (self.dims[0] + comm.size - 1) // comm.size
            d1 = min(comm.rank * mydims0, self.dims[0])
            d2 = min((comm.rank + 1) * mydims0, self.dims[0])
            mydims0 = d2 - d1
            self.mydims = (mydims0,) + self.dims[1:]
        else:
            self.mydims = ()

        fullshape = self.mydims + self.myshape

        if data is not None:
            if data.shape != fullshape:
                raise ValueError(
                    f'Bad shape for data: {data.shape} != {fullshape}')
            if data.dtype != dtype:
                raise ValueError(
                    f'Bad dtype for data: {data.dtype} != {dtype}')
        else:
            data = np.empty(fullshape, dtype)

        self.data = data
        self._matrix: Matrix | None = None

    def new(self, data=None) -> DistributedArrays:
        raise NotImplementedError

    def __getitem__(self, index):
        raise NotImplementedError

    def __iter__(self):
        for index in range(self.dims[0]):
            yield self[index]

    def flat(self):
        if self.dims == ():
            yield self
        else:
            for index in np.indices(self.dims).reshape((len(self.dims), -1)).T:
                yield self[tuple(index)]

    @property
    def matrix(self) -> Matrix:
        if self._matrix is not None:
            return self._matrix

        nx = prod(self.myshape)
        shape = (self.dims[0], prod(self.dims[1:]) * nx)
        myshape = (self.mydims[0], prod(self.mydims[1:]) * nx)
        dist = (self.comm, -1, 1)

        data = self.data.reshape(myshape)
        self._matrix = Matrix(*shape, data=data, dist=dist)

        return self._matrix

    def matrix_elements(self,
                        other,
                        *,
                        out: Matrix = None,
                        symmetric: bool = None,
                        function=None,
                        domain_sum=True,
                        cc: bool = False) -> Matrix:
        if symmetric is None:
            symmetric = self is other
        if function:
            other = function(other)
        M1 = self.matrix
        M2 = other.matrix
        out = M1.multiply(M2, opb='C', alpha=self.dv,
                          symmetric=symmetric, out=out)
        if not cc:
            out.complex_conjugate()
        # operate_and_multiply(self, self.layout.dv, out, function, ...)

        self._matrix_elements_correction(M1, M2, out, symmetric)

        if domain_sum:
            self.domain_comm.sum(out.data)

        return out

    def _matrix_elements_correction(self,
                                    M1: Matrix,
                                    M2: Matrix,
                                    out: Matrix,
                                    symmetric: bool) -> None:
        """Hook for PlaneWaveExpansion."""
        pass

    def abs_square(self,
                   weights: Array1D,
                   out: UniformGridFunctions = None) -> None:
        """Add weighted absolute square of data to output array.

        See also :xkcd:`849`.
        """
        raise NotImplementedError

    def gather(self, out=None, broadcast=False):
        raise NotImplementedError

    def interpolate(self,
                    plan1: fftw.FFTPlans = None,
                    plan2: fftw.FFTPlans = None,
                    grid: UniformGrid = None,
                    out: UniformGridFunctions = None) -> UniformGridFunctions:
        raise NotImplementedError


def operate_and_multiply(psit1, dv, out, operator, psit2):
    if psit1.comm:
        if psit2 is not None:
            assert psit2.comm is psit1.comm
        if psit1.comm.size > 1:
            out.comm = psit1.comm
            out.state = 'a sum is needed'

    comm = psit1.matrix.dist.comm
    N = len(psit1)
    n = (N + comm.size - 1) // comm.size
    mynbands = len(psit1.matrix.array)

    buf1 = psit1.new(nbands=n, dist=None)
    buf2 = psit1.new(nbands=n, dist=None)
    half = comm.size // 2
    psit = psit1.view(0, mynbands)
    if psit2 is not None:
        psit2 = psit2.view(0, mynbands)

    for r in range(half + 1):
        rrequest = None
        srequest = None

        if r < half:
            srank = (comm.rank + r + 1) % comm.size
            rrank = (comm.rank - r - 1) % comm.size
            skip = (comm.size % 2 == 0 and r == half - 1)
            n1 = min(rrank * n, N)
            n2 = min(n1 + n, N)
            if not (skip and comm.rank < half) and n2 > n1:
                rrequest = comm.receive(buf1.array[:n2 - n1], rrank, 11, False)
            if not (skip and comm.rank >= half) and len(psit1.array) > 0:
                srequest = comm.send(psit1.array, srank, 11, False)

        if r == 0:
            if operator:
                operator(psit1.array, psit2.array)
            else:
                psit2 = psit

        if not (comm.size % 2 == 0 and r == half and comm.rank < half):
            m12 = psit2.matrix_elements(psit, symmetric=(r == 0), cc=True,
                                        serial=True)
            n1 = min(((comm.rank - r) % comm.size) * n, N)
            n2 = min(n1 + n, N)
            out.array[:, n1:n2] = m12.array[:, :n2 - n1]

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        psit = buf1
        buf1, buf2 = buf2, buf1

    requests = []
    blocks = []
    nrows = (comm.size - 1) // 2
    for row in range(nrows):
        for column in range(comm.size - nrows + row, comm.size):
            if comm.rank == row:
                n1 = min(column * n, N)
                n2 = min(n1 + n, N)
                if mynbands > 0 and n2 > n1:
                    requests.append(
                        comm.send(out.array[:, n1:n2].T.conj().copy(),
                                  column, 12, False))
            elif comm.rank == column:
                n1 = min(row * n, N)
                n2 = min(n1 + n, N)
                if mynbands > 0 and n2 > n1:
                    block = np.empty((mynbands, n2 - n1), out.dtype)
                    blocks.append((n1, n2, block))
                    requests.append(comm.receive(block, row, 12, False))

    comm.waitall(requests)
    for n1, n2, block in blocks:
        out.array[:, n1:n2] = block



from math import sqrt
from ase import Atoms
from ase.dft.bandgap import bandgap
from gpaw import GPAW, FermiDirac
from gpaw.test import equal


def test_Hubbard_U(in_tmp_dir):
    """Setup up bulk NiO in an antiferromagnetic configuration."""
    # Lattice constant:
    a = 4.19
    b = a / sqrt(2)

    m = 2
    k = 2  # number of k-points

    atoms = Atoms(
        symbols='Ni2O2',
        pbc=True,
        cell=(b, b, a),
        positions=[(0, 0, 0),
                   (b / 2, b / 2, a / 2),
                   (0, 0, a / 2),
                   (b / 2, b / 2, 0)],
        magmoms=(m, -m, 0, 0))

    name = 'ni2o2'
    for setup in ['10', '10:d,6.0']:
        calc = GPAW(
            mode='pw',
            occupations=FermiDirac(width=0.05),
            setups={'Ni': setup},
            convergence={'eigenstates': 8e-4,
                         'density': 1.0e-2,
                         'energy': 0.1},
            txt=name + '.txt',
            kpts=(k, k, k),
            xc='oldPBE')
        atoms.calc = calc
        atoms.get_potential_energy()
        gap, _, _ = bandgap(calc)
        print(name, gap)
        if name == 'ni2o2':
            equal(gap, 0.8, 0.1)
        else:
            equal(gap, 4.7, 0.2)
        name += '+U'



def test_ase_features_wannierk(in_tmp_dir):
    'Test ase.dft.wannier module with k-points.'
    from ase.build import bulk
    from ase.dft.wannier import Wannier

    from gpaw import GPAW, Mixer
    from gpaw.mpi import world, serial_comm

    si = bulk('Si', 'diamond', a=5.43)

    k = 4
    if 1:
        si.calc = GPAW(kpts=(k, k, k),
                       parallel=dict(augment_grids=True),
                       mixer=Mixer(0.8, 7, 50.0),
                       txt='Si-ibz.txt')
        e1 = si.get_potential_energy()
        si.calc.write('Si-ibz.gpw', mode='all')
        si.calc.set(symmetry={'point_group': False, 'time_reversal': False},
                    txt='Si-bz.txt')
        e2 = si.get_potential_energy()
        si.calc.write('Si-bz.gpw', mode='all')
        print((e1, e2))

    def wan(calc):
        centers = [([0.125, 0.125, 0.125], 0, 1.5),
                   ([0.125, 0.625, 0.125], 0, 1.5),
                   ([0.125, 0.125, 0.625], 0, 1.5),
                   ([0.625, 0.125, 0.125], 0, 1.5)]
        w = Wannier(4, calc,
                    nbands=4,
                    # log=print,
                    initialwannier=centers)
        w.localize()
        x = w.get_functional_value()
        centers = (w.get_centers(1) * k) % 1
        c = (centers - 0.125) * 2
        print(w.get_radii())  # broken! XXX
        assert abs(c.round() - c).max() < 0.03
        c = sorted(c.round().astype(int).tolist())
        assert c == [[0, 0, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]]
        if 0:
            from ase.visualize import view
            from ase import Atoms
            watoms = calc.atoms + Atoms(symbols='X4',
                                        scaled_positions=centers,
                                        cell=calc.atoms.cell)
            view(watoms)
        return x

    calc1 = GPAW('Si-bz.gpw', txt=None, communicator=serial_comm)
    x1 = wan(calc1)
    calc2 = GPAW('Si-ibz.gpw', txt=None, communicator=serial_comm)
    calc2.wfs.ibz2bz(si)
    x2 = wan(calc2)
    if world.rank == 0:
        print((x1, x2))
    assert abs(x1 - x2) < 0.001
    assert abs(x1 - 9.71) < 0.01

    world.barrier()



# web-page: Ec_rpa.png
from myqueue.workflow import run


def workflow():
    with run(script='si.groundstate.py'):
        with run(script='si.range_rpa.py', cores=8, tmax='30m'):
            run(script='si.compare.py')
            run(script='plot_ec.py')
